# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from __future__ import annotations

import logging
from datetime import datetime
from random import randint

from sqlalchemy import Integer, String, select
from sqlalchemy.dialects.sqlite import DATETIME, JSON
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import (
    Mapped,
    mapped_column,
)
from sqlalchemy.orm import Session as ormSession

from radiotomate.db import Base

_log = logging.getLogger(__name__)


class AutoDJSlot(Base):
    """
    This entity represents one slot in the auto-DJ's schedule.

    This is a weekly schedule, accurate to the minute. Each slot defines the auto-DJ's
    constraints during its `day_of_week` (monday is 0) starting from its n-th `minute`.
    Midnight is `minute = 0`, noon is `minute = 720`, and there is at most one slot per
    day of week and per minute.

    The application assumes that the schedule's coverage is always complete, ie. there
    is always at least one row per day-of-week having `minute = 0` - those are the only
    entries whose `minute` column cannot be edited.

    Color and title are display options, that the user can set to improve its calendar's
    readability.
    """

    __tablename__ = "autodj_slots"
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    day_of_week: Mapped[int] = mapped_column(Integer, nullable=False)
    minute: Mapped[int] = mapped_column(Integer, nullable=False)
    constraints: Mapped[dict] = mapped_column(
        MutableDict.as_mutable(JSON),
        nullable=False,
        default={},
    )
    color: Mapped[str] = mapped_column(String, nullable=False, default="")
    title: Mapped[str] = mapped_column(String, nullable=False, default="")
    created: Mapped[datetime] = mapped_column(
        DATETIME, default=datetime.now, nullable=False
    )
    modified: Mapped[datetime] = mapped_column(
        DATETIME,
        nullable=False,
        default=datetime.now,
        onupdate=datetime.now,
    )

    @classmethod
    async def all(cls, session: ormSession) -> list[AutoDJSlot]:
        """
        Return all slots, sorted by day of week and minute
        """
        return await session.scalars(
            select(AutoDJSlot).order_by(AutoDJSlot.day_of_week, AutoDJSlot.minute)
        )

    @classmethod
    async def from_id(cls, session: ormSession, slot_id: int) -> AutoDJSlot | None:
        return await session.scalar(select(AutoDJSlot).filter(AutoDJSlot.id == slot_id))

    @classmethod
    async def from_time(
        cls,
        session: ormSession,
        minute: int,
        day_of_week: int | None = None,
        exact=False,
    ) -> AutoDJSlot | None:
        """
        Return the first slot that matches given time constraints, if any.

        If ``day_of_week`` is not provided, search any day that matches given time.

        If ``exact`` is ``False`` (by default), search for the current slot, ie. the
        first one having given ``minute`` or less.
        """
        q = select(AutoDJSlot)
        if exact:
            q = q.filter(AutoDJSlot.minute == minute)
        else:
            q = q.filter(AutoDJSlot.minute <= minute)
        if day_of_week is not None:
            q = q.filter(AutoDJSlot.day_of_week == day_of_week)
        q = q.order_by(AutoDJSlot.minute.desc())
        return await session.scalar(q)

    @classmethod
    async def current_filter(cls, session: ormSession) -> str:
        """
        Return a Beets expression matching the AutoDJ constraint at current time

        Returns an empty stringif something goes wrong.
        """
        now = datetime.now()
        minute = now.minute + 60 * now.hour
        slot = await cls.from_time(session, minute, now.weekday())
        if slot:
            return slot.pick_filter()
        else:
            _log.warning("no AutoDJSlot found !!")
            return ""

    def pick_filter(self) -> str:
        cumulated = 0
        max_weight = self.constraints.get("totalweight", 0)
        filters = self.constraints.get("filters", [])
        if not filters:
            return ""
        weight = randint(0, max_weight - 1)
        for f in filters:
            cumulated += f[0]
            if len(f) == 2 and weight < cumulated:
                return f[1]
        _log.warning(
            "Consistency error: picked weight %r, did not match any in %r",
            weight,
            filters,
        )
        return ""

    @property
    def display_time(self) -> str:
        hour = self.minute // 60
        minute = self.minute % 60
        return f"{hour:02d}:{minute:02d}"



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

# ruff: noqa: E402 - not all imports are at top of file because of the "responses" mock
"""
This module contains a mock Beets library, that should be used only for tests and demo
mode.
"""

import logging
import sys
from asyncio import Queue
from unittest.mock import MagicMock

from radiotomate.models.sound import Sound

# we mock the `responses` module entirely, it's imported by Beets' TestHelper but not
# used in our case
sys.modules["responses"] = MagicMock()

from beets.test.helper import TestHelper

from radiotomate.beets import BeetsIntegration
from radiotomate.scheduler_api import artistic_generator

# Test library will contain 6 tracks of this artist.
FAKE_ARTIST = "Camille Tester"


class BeetsMockIntegration(BeetsIntegration):
    BACKGROUND_ANALYSIS_DELAY = 0.0

    def __init__(self):
        """
        As opposed to :py:class:`radiotomate.beets.BeetsIntegration`, this constructor
        can be called directly. However, please call :py:meth:`teardown` before closing
        the application. Note that it does not call its super constructor!

        We re-use Beets' own fixture, but also pre-fill with some test content.

        The relay gain analysis goes trough our background task, but does not really
        call the ReplayGainPlugin.
        """
        TestHelper.db_on_disk = True
        helper = TestHelper()
        helper.setup_beets()
        logging.getLogger("beets").setLevel(logging.INFO)  # defaults to verbose
        self.helper = helper
        self.lib = helper.lib
        self.rg = None
        self.analyzer_q = Queue()

        for _ in range(5):
            helper.add_item(artist=FAKE_ARTIST, title=artistic_generator())

        for _ in range(100):
            helper.add_item(artist=artistic_generator(), title=artistic_generator())

    def teardown(self):
        self.helper.teardown_beets()

    async def _do_analyze_rg(self, sound: Sound):
        sound.gain = -1.0
        sound.peak = -1.0



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from __future__ import annotations

import logging

from sqlalchemy import String, select
from sqlalchemy.orm import Mapped, Session, mapped_column

from radiotomate.db import Base

_log = logging.getLogger(__name__)


class Setting(Base):
    """
    In-dabatase settings.

    Those settings can be edited from the web interface by
    admins. Any setting that would not require a webapp restart can be stored
    here, otherwise it should be in the YAML file. However, they will be loaded
    in the same dict as those in the YAML file, and they might override them!

    Also note that sub-dicts are created when the key contains a dot.
    """

    __tablename__ = "settings"
    key: Mapped[str] = mapped_column(String, primary_key=True, nullable=False)
    value: Mapped[str] = mapped_column(String, nullable=False)

    @classmethod
    async def from_key(cls, session: Session, key: str) -> Setting | None:
        return await session.scalar(select(Setting).filter(Setting.key == key))

    @classmethod
    async def load_all(cls, session: Session) -> dict:
        settings = {}
        rows = await session.scalars(select(Setting))
        for setting in rows:
            settings[setting.key] = setting.value
        return settings



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import logging

from quart import Blueprint

from radiotomate.scheduler.watchdog import Watchdog
from radiotomate.version import get_version

_log = logging.getLogger(__name__)

blueprint = Blueprint("version", __name__)


@blueprint.get("/version")
async def version():
    return {
        "radiotomate": get_version(),
        "liquidsoap": Watchdog.liquidsoap_version,
    }



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import logging

from quart import Blueprint, g, render_template, request, url_for
from werkzeug.exceptions import BadRequest, Conflict, NotFound

from radiotomate.auth import permission_required
from radiotomate.models import User

_log = logging.getLogger(__name__)

blueprint = Blueprint("users", __name__, template_folder="templates")


@blueprint.get("/users")
@permission_required("admin")
async def index():
    users = await User.all(g.dbsession)
    return await render_template("users/index.jinja", users=users)


@blueprint.get("/users/add")
@permission_required("admin")
async def add_form():
    return await render_template("users/user_form.jinja")


@blueprint.post("/users")
@permission_required("admin")
async def add():
    form = await request.form
    username = form.get("username")
    if username:
        username = username.strip()

    if not username:
        raise BadRequest("Please provide an username")
    if not form.get("password"):
        raise BadRequest("Please provide a password")

    if await User.from_username(g.dbsession, username):
        raise Conflict(f"User {username} already exists")

    notes = form.get("notes")
    user = User(username=username, notes=notes)
    user.update_password(form["password"])
    user.update_roles(form)

    g.dbsession.add(user)
    await g.dbsession.commit()
    return (
        "",
        200,
        {
            "HX-Redirect": url_for("users.index"),
        },
    )


@blueprint.get("/users/<int:user_id>")
@permission_required("admin")
async def get(user_id: int):
    user = await User.from_id(g.dbsession, user_id)
    if not user:
        raise NotFound(f"User {user_id} not found")
    return await render_template("users/user_form.jinja", user=user)


@blueprint.put("/users/<int:user_id>")
@permission_required("admin")
async def edit(user_id: int):
    form = await request.form
    user = await User.from_id(g.dbsession, user_id)
    if not user:
        raise NotFound(f"User {user_id} not found")

    username = form.get("username")
    if username:
        username = username.strip()
    if not username:
        raise BadRequest("Please provide an username")
    if username != user.username and await User.from_username(g.dbsession, username):
        raise Conflict(f"User {username} already exists")
    user.username = username

    user.notes = form.get("notes")
    if form.get("password"):
        user.update_password(form["password"])

    user.update_roles(form)

    await g.dbsession.commit()
    return (
        "",
        200,
        {
            "HX-Redirect": url_for("users.index"),
        },
    )


@blueprint.delete("/users/<int:user_id>")
@permission_required("admin")
async def delete(user_id: int):
    user = await User.from_id(g.dbsession, user_id)
    if not user:
        raise NotFound(f"User {user_id} not found")
    await g.dbsession.delete(user)
    await g.dbsession.commit()
    return ""



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import logging
from asyncio import Queue
from datetime import datetime
from http.client import HTTPException

from hypercorn.utils import ShutdownError
from quart import Blueprint, make_response, request

from radiotomate.auth import login_required, permission_required
from radiotomate.quart import or_shutdown
from radiotomate.scheduler_api import Scheduler
from radiotomate.templates import render_macro as general_render_macro

_log = logging.getLogger(__name__)

blueprint = Blueprint("live", __name__, template_folder="templates")
watching_clients = set()


async def render_macro(macro, **kwargs) -> str:
    return await general_render_macro("live/macros.jinja", macro, **kwargs)


@blueprint.get("/live")
@login_required
async def get():
    if "text/event-stream" not in request.accept_mimetypes:
        raise HTTPException(400, "text/event-stream not supported")

    queue = Queue()
    watching_clients.add(queue)

    async def send_now_playing():
        try:
            while True:
                data = await or_shutdown(queue.get())
                yield f"event: message\ndata: {data}\n\n".encode()
        except ShutdownError:
            pass
        except Exception as exc:
            _log.exception(exc)
        finally:
            watching_clients.remove(queue)
            _log.debug(
                "shutting down a /live stream - %d remaining",
                len(watching_clients),
            )

    response = await make_response(
        send_now_playing(),
        {
            "Content-Type": "text/event-stream",
            "Cache-Control": "no-cache",
            "Transfer-Encoding": "chunked",
        },
    )
    response.timeout = None
    return response


async def watch_livedata_task():
    async for md in Scheduler.get().live():
        try:
            remaining = round(float(md.get("remaining", 0.0)))
        except OverflowError:  # liquidsoap may send "inf"
            remaining = 0
        if md.get("elapsed"):
            md["elapsed"] = round(float(md.get("elapsed")))
        if md.get("time"):
            time_dt = datetime.fromisoformat(md.get("time"))
            md["date"] = time_dt.strftime("%d/%m/%Y")
            md["time"] = time_dt.strftime("%H:%M:%S")
        rendered = await render_macro("nowplayling", data=md, remaining=remaining)
        rendered = rendered.replace("\n", " ")  # \n\n is a separator for SSE
        for queue in list(watching_clients):
            await queue.put(rendered)


@blueprint.delete("/live")
@permission_required("live")
async def skip():
    await (Scheduler.get()).skip()
    return "", 200



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

# load sub-commands
from radiotomate.commands import (  # noqa: F401
    install,
    interface,
    radiotomate_cli,
    scheduler,
    update,
    users,
)

if __name__ == "__main__":
    radiotomate_cli()



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
The interface's ASGI app factory
"""

import json
import logging
from importlib.metadata import version
from pathlib import Path
from traceback import extract_tb

from quart import current_app, request
from quart_auth import QuartAuth, Unauthorized, current_user
from werkzeug.exceptions import HTTPException

import radiotomate.models
from radiotomate.auth import RadiotomateAuth, redirect_to_login
from radiotomate.beets import BeetsIntegration
from radiotomate.db import PathLike, QuartAlchemy
from radiotomate.interface import (
    autodj,
    carts,
    history,
    home,
    library,
    live,
    login,
    users,
)
from radiotomate.quart import CustomQuart
from radiotomate.templates import RADIOTOMATE_HELPERS

_log = logging.getLogger(__name__)


async def htmx_errors_as_messages(exc: Exception):
    if isinstance(exc, HTTPException):
        radiotomate_message = {
            "text": exc.description,
            "icon": "warning",
        }
    else:
        # At this point we're out of the exception handling context, so the usual
        # log.exception() won't show the stack.
        message = (
            "".join(extract_tb(exc.__traceback__).format())
            + f"{exc.__class__.__name__}: {exc}"
        )
        _log.error(message)
        if not current_app.debug:
            message = "Please check the logs for more information"
        radiotomate_message = {
            "title": "Unhandled exception",
            "text": message,
            "icon": "error",
        }
        exc.code = 500

    if request.headers.get("HX-Request") == "true":
        headers = {
            "HX-Trigger": json.dumps({"radiotomateMessage": radiotomate_message}),
        }
        return (radiotomate_message["text"], exc.code, headers)
    else:
        return (f"<pre>{radiotomate_message['text']}</pre>", exc.code)


def app_factory(config: dict, reload: bool, beets: BeetsIntegration):
    app = CustomQuart(__name__)

    app.config.from_mapping(
        {
            "RADIOTOMATE_VERSION": version("radiotomate"),
            "SECRET_KEY": config["interface"]["cookie_salt"],
            "TEMPLATES_AUTO_RELOAD": reload,
            "ALCHEMY_ENGINE_CONFIG": config["db"],
            "QUART_AUTH_COOKIE_NAME": "session",
            "QUART_AUTH_COOKIE_SAMESITE": "Strict",
            "QUART_AUTH_COOKIE_SECURE": False,
            "QUART_AUTH_SALT": config["interface"]["cookie_salt"],
            "MAX_CONTENT_LENGTH": 10 * 1024 * 1024 * 1024,  # 10GB,
            "DATA_ROOT": Path(config["data"]["root"]),
            "PLAYOUT_TOKEN": config["playout_process_config"]["token"],
        },
    )

    app.register_blueprint(home.blueprint)
    app.register_blueprint(users.blueprint)
    app.register_blueprint(login.blueprint)
    app.register_blueprint(carts.blueprint)
    app.register_blueprint(live.blueprint)
    app.register_blueprint(autodj.blueprint)
    app.register_blueprint(library.blueprint)
    app.register_blueprint(history.blueprint)

    app.jinja_env.globals.update(**RADIOTOMATE_HELPERS)
    app.register_error_handler(Exception, htmx_errors_as_messages)
    app.register_error_handler(Unauthorized, redirect_to_login)

    PathLike.base_path = app.config["DATA_ROOT"]
    db = QuartAlchemy(app)

    beets.init_app(app)

    @app.before_serving
    async def on_start():
        # load settings
        async with db.session() as session:
            dbsettings = await radiotomate.models.Setting.load_all(session)
            app.config.from_mapping(dbsettings)

        # start the live data watch
        app.add_background_task(live.watch_livedata_task)

    auth_manager = QuartAuth()
    auth_manager.user_class = RadiotomateAuth
    auth_manager.init_app(app)

    @app.before_request
    @app.before_websocket
    async def preload_user():
        """
        Preloads a complete Session and User object.

        Except when serving from the static endpoint, because it does not open
        a DB session.
        """
        if request.endpoint != "static":
            read_only = request.method in ("GET", "HEAD", "OPTIONS")
            await current_user.preload_attributes(update_latest_action=not read_only)

    return app



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from __future__ import annotations

import logging
from datetime import datetime

from argon2 import PasswordHasher
from argon2.exceptions import VerifyMismatchError
from sqlalchemy import Integer, String, delete, select
from sqlalchemy.dialects.sqlite import DATETIME, JSON
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.orm import Session as ormSession
from werkzeug.datastructures import MultiDict  # noqa: TC002

from radiotomate.db import Base
from radiotomate.enums import Role

_log = logging.getLogger(__name__)


class User(Base):
    """
    User info and roles
    """

    __tablename__ = "users"
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    username: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    password: Mapped[str] = mapped_column(String, nullable=False)
    notes: Mapped[str] = mapped_column(String, nullable=False, default="")
    created: Mapped[datetime] = mapped_column(DATETIME, default=datetime.now)
    modified: Mapped[datetime] = mapped_column(
        DATETIME,
        default=datetime.now,
        onupdate=datetime.now,
    )
    roles: Mapped[dict] = mapped_column(
        MutableDict.as_mutable(JSON),
        default={},
    )  # https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#sqlalchemy.dialects.sqlite.JSON

    sessions: Mapped[list[Session]] = relationship(  # noqa: F821
        lazy=True,
        back_populates="user",
        cascade="save-update, merge, expunge, delete",
    )

    @classmethod
    async def all(cls, session: ormSession) -> list[User]:
        return await session.scalars(select(User).order_by(User.username))

    @classmethod
    async def from_username(cls, session: ormSession, username: str) -> User | None:
        return await session.scalar(select(User).filter(User.username == username))

    @classmethod
    async def from_id(cls, session: ormSession, user_id: int) -> User | None:
        return await session.scalar(select(User).filter(User.id == user_id))

    @classmethod
    async def check(
        cls,
        session: ormSession,
        username: str,
        password: str,
    ) -> User | None:
        user = await cls.from_username(session, username)
        if user:
            ph = PasswordHasher()
            try:
                if ph.verify(user.password, password):
                    if ph.check_needs_rehash(user.password):
                        user.password = ph.hash(password)
                        await session.flush()
                    return user
            except VerifyMismatchError:
                return None
            except Exception as e:
                _log.exception(e)
        return None

    @classmethod
    async def delete(cls, session: ormSession, user_id: int):
        await session.execute(delete(User).where(User.id == user_id))

    def update_password(self, new_password):
        ph = PasswordHasher()
        self.password = ph.hash(new_password)

    def update_roles(self, form: MultiDict):
        roles = {}
        for role in Role:
            if form.get(f"can_{role}") == "true":
                roles[role] = True
        self.roles = roles

    def can_admin(self) -> bool:
        return self.roles.get("admin", False)

    def can_stream(self) -> bool:
        return self.can_admin() or self.roles.get("stream", False)

    def can_live(self) -> bool:
        return self.can_admin() or self.roles.get("live", False)

    def can_carts(self) -> bool:
        return self.can_admin() or self.roles.get("carts", False)

    def can_not_audio(self) -> bool:
        return self.roles.get("not_audio", False)

    def can_autodj(self) -> bool:
        return self.can_admin() or self.roles.get("autodj", False)


class LoggedOutUser:
    """
    This object fills current_user.user when the user is not authenticated.
    This simplifies authorization checks by providing all ``can_*()`` methods
    """

    def __init__(self):
        self.id = None
        self.username = None
        self.roles = {}

    def can_admin(self) -> bool:
        return False

    def can_stream(self) -> bool:
        return False

    def can_live(self) -> bool:
        return False

    def can_carts(self) -> bool:
        return False

    def can_not_audio(self) -> bool:
        return False

    def can_autodj(self) -> bool:
        return False



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import click

from . import radiotomate_cli


@radiotomate_cli.command()
@click.option("--demo", is_flag=True, help="Runs in demo mode")
@click.option(
    "--reload",
    is_flag=True,
    help="Reloads app on changes (when developing))",
)
@click.pass_obj
def interface(config_dict, demo, reload):
    """
    Start the Radiotomate web application
    """
    import asyncio

    from hypercorn.asyncio import serve

    from radiotomate.interface_app import app_factory
    from radiotomate.quart import CustomQuart, ShutdownManager
    from radiotomate.scheduler_api import Scheduler

    if demo:
        from radiotomate.beets.demo import BeetsMockIntegration

        beets = BeetsMockIntegration()
    else:
        from radiotomate.beets import BeetsIntegration

        beets = BeetsIntegration.default()

    app = app_factory(config_dict, reload, beets)
    Scheduler.init(config_dict, app, demo=demo)
    if reload:
        app.run(use_reloader=reload, host="0.0.0.0", port=6811)
    else:

        @app.before_serving
        async def on_start():
            ShutdownManager().setup_signal_handler()

        config = CustomQuart.make_basic_hypercorn_config()
        config.bind = ["0.0.0.0:6811"]
        asyncio.run(serve(app, config, shutdown_trigger=ShutdownManager().wait))

    if demo:
        beets.teardown()



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import secrets
from pathlib import Path
from shutil import copy

import click

YAML_TEMPLATE = """
---

playout_process_config:
    outputs:
        #- driver: pulseaudio
        #  # empty(null) will let pulseaudio choose the default device.
        #  device:

        #- driver: icecast
        #  encoder: vorbisvbr
        #  quality: 0.3
        #  host: "127.0.0.1"
        #  port: 9876
        #  mount: "radio.ogg"
        #  password: "secret"
    jingles_delay: 780.
    input_name: "{http_input_name}"
    input_min_buffer: 5.
    input_max_buffer: 30.
    start_sound: {start_sound_path}
    {playout_log_config}
    token: {playout_token}

metadata_log:
    # list of metadata fields that should be stored when available, in addition to
    # artist/title/album.
    extra_fields:
        - genre
        - year

    # if you want to POST metadata to some other websites, remove leading # below:
    #relay_to:
    #    - url: "https://website.radio/playlist/add_item"
    #      # additional POST fields
    #      add_field:
    #          SECRET_KEY: "secret value"
    #          SOURCE_NAME: "radiotomate"
    #    - url: "https://myradio.org/track_recorder.php"
    #      # and/or additional POST headers
    #      add_header:
    #          Authorization: "Basic YWxhZGRpbjpvcGVuc2VzYW1l"

data:
    root: {data_root}

db:
    url: "sqlite+aiosqlite:///{db}"

interface:
    # should be set randomly when installing. Changing it would log out everyone.
    cookie_salt: {cookie_salt}


############# Logging configuration ##########
logging:
    version: 1
    disable_existing_loggers: false

    formatters:
        generic:
            format: "%(asctime)s %(levelname)-5.5s [%(name)s:%(lineno)s] %(message)s"

    handlers:
        default:
            formatter: generic
            class: "logging.StreamHandler"
        interface:
            formatter: generic
            {interface_handler}
        scheduler:
            formatter: generic
            {scheduler_handler}

    root:
        level: INFO
        # load_config() will add one handler (interface or scheduler or default)
        # this allows interface and scheduler to user the same configuration while
        # outputting to their own file in production.
        handlers: []

    loggers:
        # This allows radiotomate to log@DEBUG while keeping others >= INFO
        radiotomate:
            level: {log_level}
            handlers: []
            propagate: false
            qualname: radiotomate

        # set level to INFO if you need to see requests made by Python processes
        httpx:
            level: WARNING
            qualname: httpx
"""

STARTWAV = "default_start_sound.wav"


class Installer:
    def __init__(self, path_data: Path):
        self.path_data = path_data
        self.http_input_name = "stream"

    @property
    def path_yaml(self):
        return self.path_data / "radiotomate.yaml"

    @property
    def path_db(self):
        return self.path_data / "radiotomate.db"

    @property
    def path_interface_log(self):
        return self.path_data / "interface.log"

    @property
    def path_scheduler_log(self):
        return self.path_data / "scheduler.log"

    @property
    def path_playout_log(self):
        return self.path_data / "playout.log"

    def _potential_paths(self):
        yield from [
            self.path_yaml,
            self.path_db,
            self.path_interface_log,
            self.path_scheduler_log,
            self.path_playout_log,
            self.path_data / STARTWAV,
        ]

    def instance_exists(self) -> bool:
        return any(path.exists() for path in self._potential_paths())

    def revert(self):
        for path in self._potential_paths():
            if path.exists():
                path.unlink()

    def install(self, dev=False):
        if dev:
            interface_handler = 'class: "rich.logging.RichHandler"'
            scheduler_handler = 'class: "rich.logging.RichHandler"'
            log_level = "DEBUG"
            playout_log_config = ""
        else:
            playout_log_config = f'log_path: "{self.path_playout_log}"'
            log_level = "INFO"
            scheduler_handler = f"""class: "logging.handlers.RotatingFileHandler"
            filename: "{self.path_scheduler_log}"
            maxBytes: 10000000
            backupCount: 10
            """
            interface_handler = f"""class: "logging.handlers.RotatingFileHandler"
            filename: "{self.path_interface_log}"
            maxBytes: 10000000
            backupCount: 10
            """

        click.echo(f"Initializing data folder: {self.path_data}")
        (self.path_data / "carts").mkdir(parents=True, exist_ok=True)
        start_sound = copy(Path(__file__).parent / STARTWAV, self.path_data / STARTWAV)

        with self.path_yaml.open("w") as yaml_file:
            click.echo(f"Writing configuration file: {self.path_yaml}")
            yaml_file.write(
                YAML_TEMPLATE.format(
                    data_root=self.path_data,
                    db=self.path_db,
                    http_input_name=self.http_input_name,
                    interface_handler=interface_handler,
                    scheduler_handler=scheduler_handler,
                    log_level=log_level,
                    playout_log_config=playout_log_config,
                    cookie_salt=secrets.token_hex(),
                    playout_token=secrets.token_hex(),
                    start_sound_path=str(start_sound),
                ),
            )

    def recap(self):
        click.secho("\nAll done !", fg="green", bold=True)
        click.echo("Radiotomate configuration, database, logs and files will be in:")
        click.secho(str(self.path_data), bold=True)
        click.secho("\nWe advise you backup regularly the content of this folder.")



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
This modules contains our extensions to Quart.

Our web applications are running some permanent tasks in the background,
so we need to customize how to shutdown the whole application.
"""

import asyncio
import logging
import signal
from collections.abc import Awaitable, Callable, Coroutine

from hypercorn import Config
from hypercorn.asyncio import serve
from hypercorn.utils import ShutdownError
from quart import Quart

_log = logging.getLogger(__name__)


class ShutdownManager:
    """
    We have to handle interruptions/signals ourselves to implement ``or_shutdown``,
    and per [Hypercorn documentation](https://hypercorn.readthedocs.io/en/latest/how_to_guides/api_usage.html#graceful-shutdown)
    """

    _instance = None
    _task = None
    _event = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._task = None
            cls._instance._event = asyncio.Event()
        return cls._instance

    @property
    def event(self):
        return self._event

    async def wait(self):
        await self._event.wait()

    @property
    def task(self):
        if self._task is None:
            self._task = asyncio.create_task(
                self._event.wait(),
                name="RadiotomateShutdownManager",
            )
        return self._task

    def trigger_shutdown(self, *_):
        self._event.set()

    def setup_signal_handler(self):
        """
        This is copied from Hypercorn's trigger setup, but with our own Event.
        """
        loop = asyncio.get_event_loop()
        for signal_name in ("SIGINT", "SIGTERM", "SIGBREAK"):
            if hasattr(signal, signal_name):
                try:
                    loop.add_signal_handler(
                        getattr(signal, signal_name),
                        self.trigger_shutdown,
                    )
                except NotImplementedError:
                    # Add signal handler may not be implemented on Windows
                    signal.signal(getattr(signal, signal_name), self.trigger_shutdown)


async def or_shutdown(coro: Coroutine):
    """
    Awaits for ``coro`` and return its result, or raises ``ShutdownError`` if a
    shutdown signal was received.
    """
    coro_task = asyncio.create_task(coro)
    try:
        await asyncio.wait(
            (coro_task, ShutdownManager().task),
            return_when=asyncio.FIRST_COMPLETED,
        )
        if coro_task.done():
            if exn := coro_task.exception():
                raise exn
            return coro_task.result()
    finally:
        if not coro_task.done():
            coro_task.cancel()
            raise ShutdownError()


class CustomQuart(Quart):
    jinja_options = {  # noqa: RUF012
        "autoescape": True,
    }

    def run_task(  # noqa: PLR0913
        self,
        host: str = "127.0.0.1",
        port: int = 5000,
        debug: bool | None = None,
        ca_certs: str | None = None,
        certfile: str | None = None,
        keyfile: str | None = None,
        shutdown_trigger: Callable[..., Awaitable[None]] | None = None,
    ) -> Coroutine[None, None, None]:
        """
        We need to override Quart's ``run_task`` because it's great to have its
        reloader, but the original method hardcodes a few things in the configuration.

        We also need to plug ShutdownManager behind ``shutdown_trigger``
        because the provided trigger will also stop the reloader. That's why
        when reloading we don't need to call
        ``ShutdownManager().setup_signal_handler()``.
        """
        config = self.make_basic_hypercorn_config()
        config.bind = [f"{host}:{port}"]
        config.ca_certs = ca_certs
        config.certfile = certfile
        config.keyfile = keyfile
        if debug is not None:
            self.debug = debug

        async def wrap_trigger():
            await shutdown_trigger()
            ShutdownManager().trigger_shutdown()

        return serve(self, config, shutdown_trigger=wrap_trigger)

    @classmethod
    def make_basic_hypercorn_config(cls) -> Config:
        config = Config()
        config.access_log_format = "%(h)s %(r)s %(s)s %(b)s %(L)s"
        config.accesslog = logging.getLogger("hypercorn.access")
        config.errorlog = config.accesslog
        return config



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from __future__ import annotations

import logging
from datetime import datetime
from pathlib import Path  # noqa: TC003

from sqlalchemy import Boolean, ForeignKey, Integer, String, func, or_, select
from sqlalchemy.dialects.sqlite import DATETIME
from sqlalchemy.orm import Mapped, Session, mapped_column, relationship

from radiotomate.db import Base

_log = logging.getLogger(__name__)


class Sound(Base):
    """
    A sound file in a cart.

    Note that path is relative to the app's data root, but absolute when instanciated
    in Python. This is provided by `radiotomate.db`.
    """

    __tablename__ = "sounds"
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    cart_id: Mapped[int] = mapped_column(
        Integer,
        ForeignKey("carts.id"),
        nullable=False,
        index=True,
    )
    rank: Mapped[int] = mapped_column(Integer, default=1)
    path: Mapped[Path]
    duration: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    title: Mapped[str] = mapped_column(String, nullable=False)
    active: Mapped[bool] = mapped_column(Boolean, nullable=False, default=True)
    created: Mapped[datetime] = mapped_column(
        DATETIME,
        default=datetime.now,
        nullable=False,
    )
    last_played: Mapped[datetime] = mapped_column(DATETIME, nullable=True)
    gain: Mapped[float]
    peak: Mapped[float]
    uploader_id: Mapped[int] = mapped_column(Integer, ForeignKey("users.id"))

    cart: Mapped[Cart] = relationship(lazy="joined", back_populates="sounds")  # noqa: F821

    uploader: Mapped[User] = relationship()  # noqa: F821

    @property
    def available(self) -> bool:
        """
        The active field is explicitely set by the user, while gain should be computed
        soon by our scheduler's analyzer.

        Note that this logic is also implemented for Cart's query expressions.
        """
        return self.active and self.gain is not None

    @classmethod
    async def from_id(cls, session: Session, sound_id: int) -> Sound | None:
        return await session.scalar(select(Sound).filter(Sound.id == sound_id))

    @classmethod
    async def next_rank(cls, session: Session, cart_id: int) -> int:
        """
        Returns the next rank for a cart's sounds.
        """
        q = select(func.max(Sound.rank).filter(Sound.cart_id == cart_id))
        result = await session.scalar(q)
        if result is None:
            return 1
        else:
            return result + 1

    @classmethod
    async def update_ranks(cls, session: Session, cart_id: int) -> list[Sound]:
        """
        Updates the ranks of all sounds in a cart, ensuring they're numbered from 1.
        Returns all sounds.
        """
        q = select(Sound).filter(Sound.cart_id == cart_id).order_by(Sound.rank)
        sounds = list(await session.scalars(q))
        rank = 1
        for sound in sounds:
            if sound.rank != rank:
                sound.rank = rank
            rank += 1
        return sounds

    @classmethod
    async def change_path_prefix(
        cls,
        session: Session,
        cart_id: int,
        old: Path,
        new: Path,
    ):
        """
        Updates path when ``cart_id`` is moved from ``old`` to ``new``.
        """
        q = select(Sound).filter(Sound.cart_id == cart_id)
        sounds = list(await session.scalars(q))
        for sound in sounds:
            sound.path = new / sound.path.relative_to(old)

    @classmethod
    async def ids_without_gain(cls, session: Session) -> list[int]:
        """
        Returns the list of Sound.id that are missing replaingain fields
        """
        q = select(Sound.id).filter(or_(Sound.gain == None, Sound.peak == None))  # noqa: E711
        return list(await session.scalars(q))



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import logging
import shutil
from urllib.parse import urlparse

import mutagen
from quart import (
    Blueprint,
    current_app,
    g,
    render_template,
    request,
    send_file,
    url_for,
)
from werkzeug.exceptions import BadRequest, Conflict, NotFound

from radiotomate.auth import current_user, login_required, permission_required
from radiotomate.enums import CartMode, DayNames, ScheduleMode
from radiotomate.interface import safe_int, safe_path, stream_form
from radiotomate.models import Cart, Sound
from radiotomate.models.cart import URL_TO_AUTODJ_QUEUE
from radiotomate.scheduler_api import Scheduler
from radiotomate.templates import render_macro as general_render_macro

_log = logging.getLogger(__name__)

blueprint = Blueprint("carts", __name__, template_folder="templates")


async def render_macro(macro, **kwargs) -> str:
    return await general_render_macro("carts/macros.jinja", macro, **kwargs)


async def poke_scheduler(cart_id: int) -> None:
    await Scheduler.get().update_schedule(cart_id)


def apply_schedule(cart: Cart, form: dict) -> Cart:
    """
    Transform form data ("cron formula" fields) into a cart's schedule dict.
    We PUT it to the scheduler process, that may invalidate the dict. At least, it will
    cancel any existing schedule.

    This returns the Cart to allow chaining.
    """
    try:
        cart.schedule_mode = ScheduleMode[form.get("schedule")]
    except ValueError:
        raise BadRequest(f"schedule_mode should be in {list(ScheduleMode)}") from None

    def get_or(field, default):
        received = form.get(field)
        if received:
            received = received.strip()
        if received:
            return received
        else:
            return default

    if cart.schedule_mode is ScheduleMode.JINGLES:
        cart.schedule_minute = get_or("minute", "*")
        cart.schedule_second = get_or("second", "*")
    else:
        cart.schedule_minute = get_or("minute", "0")
        cart.schedule_second = get_or("second", "0")

    cart.schedule_year = get_or("year", "*")
    cart.schedule_month = get_or("month", "*")
    cart.schedule_day = get_or("day", "*")
    cart.schedule_week = get_or("week", "*")
    cart.schedule_day_of_week = get_or("day_of_week", "*")
    cart.schedule_hour = get_or("hour", "*")

    try:
        cart.to_timed_trigger()
        cart.schedule_correct = True
    except Exception as e:
        cart.schedule_correct = False
        _log.warning("incorrect schedule %s, got %r", cart.schedule_repr, e)

    return cart


def apply_max_duration(cart: Cart, form: dict) -> None:
    """
    Attempt to convert the minutes/seconds inputs into a cart's max_duration field.
    If values are not correct, the field is left as it was.
    """
    minutes = safe_int(form.get("max_duration_minutes"), 0, positive=True)
    seconds = safe_int(form.get("max_duration_seconds"), 0, positive=True)
    if minutes == 0 and seconds == 0:
        cart.max_duration = None
    else:
        seconds += 60 * minutes
        cart.max_duration = seconds
    if cart.mode is CartMode.RELAY and cart.max_duration is None:
        raise BadRequest("Stream relays should have a maximum duration")


def apply_url(cart: Cart, form: dict) -> None:
    """
    Cart.url is also used by non-relay carts, to flag if the sounds cart should push
    to the auto-dj queue (cf. the ``URL_TO_AUTODJ_QUEUE`` constant).
    """
    cart.url = None
    if cart.mode is CartMode.RELAY:
        cart.url = form.get("url")
        parsed = urlparse(cart.url)
        if not parsed.scheme or not parsed.hostname or not parsed.path:
            raise BadRequest("This does not look like a complete URL")
    else:
        flag = form.get("queue")
        if flag == URL_TO_AUTODJ_QUEUE:
            cart.url = URL_TO_AUTODJ_QUEUE


@blueprint.get("/carts")
@login_required
async def index():
    carts = await Cart.all(g.dbsession)
    return await render_template("carts/index.jinja", carts=carts)


@blueprint.get("/carts/add")
@permission_required("carts")
async def add_form():
    return await render_template(
        "carts/cart_form.jinja", is_advanced="false", day_names=DayNames
    )


@blueprint.post("/carts")
@permission_required("carts")
async def add():
    form = await request.form
    title = form.get("title")
    if title:
        title = title.strip()
    if not title:
        raise BadRequest("Please provide a title")

    if await Cart.from_title(g.dbsession, title):
        raise Conflict(f"Cart {title} already exists")

    cart = Cart(title=title, notes=form.get("notes"))

    try:
        mode = form.get("mode")
        cart.mode = CartMode[mode]
    except KeyError:
        raise BadRequest(f"Incorrect mode value: {mode}") from None
    apply_max_duration(cart, form)
    apply_schedule(cart, form)
    apply_url(cart, form)

    g.dbsession.add(cart)
    await g.dbsession.commit()

    cart_root = current_app.config["DATA_ROOT"] / "carts"
    cart.path = safe_path(cart_root, title, prefix=str(cart.id))
    cart.path.mkdir(parents=True)
    await g.dbsession.commit()

    if cart.schedule_mode is ScheduleMode.TIMED:
        # do this even if not schedule_correct because it will at least remove that
        # cart from the schedule
        current_app.add_background_task(poke_scheduler, cart.id)

    return (
        "",
        200,
        {
            "HX-Redirect": url_for("carts.index"),
        },
    )


@blueprint.get("/carts/<int:cart_id>")
@permission_required("carts")
async def edit_form(cart_id: int):
    cart = await Cart.from_id(g.dbsession, cart_id)
    if not cart:
        raise NotFound(f"Cart {cart_id} not found")
    coming_from = request.args.get("coming_from")
    return await render_template(
        "carts/cart_form.jinja",
        cart=cart,
        coming_from=coming_from,
        is_advanced=str(cart.schedule_is_advanced).lower(),
        day_names=DayNames,
    )


@blueprint.put("/carts/<int:cart_id>")
@permission_required("carts")
async def edit(cart_id: int):
    form = await request.form
    cart = await Cart.from_id(g.dbsession, cart_id)
    if not cart:
        raise NotFound(f"Cart {cart_id} not found")

    title = form.get("title")
    if title:
        title = title.strip()
    if not title:
        raise BadRequest("Please provide a title")
    if title != cart.title:
        if await Cart.from_title(g.dbsession, title):
            raise Conflict(f"Cart {title} already exists")
        cart.title = title
        cart_root = current_app.config["DATA_ROOT"] / "carts"
        new_path = safe_path(cart_root, cart.title, prefix=str(cart.id))
        if new_path != cart.path:
            cart.path.rename(new_path)
            await Sound.change_path_prefix(
                g.dbsession,
                cart.id,
                old=cart.path,
                new=new_path,
            )
            cart.path = new_path

    cart.notes = form.get("notes")

    try:
        mode = form["mode"]
        try:
            cart.mode = CartMode[mode]
        except KeyError:
            raise BadRequest(f"Incorrect mode value: {mode}") from None
    except KeyError:  # `mode` is disabled for relays, so browser don't post it
        pass

    apply_max_duration(cart, form)
    apply_schedule(cart, form)
    apply_url(cart, form)
    await g.dbsession.commit()

    if cart.schedule_mode is ScheduleMode.TIMED:
        # do this even if not schedule_correct because it will at least remove that
        # cart from the schedule
        current_app.add_background_task(poke_scheduler, cart.id)

    if form.get("coming_from") == "sounds":
        redirect_to = url_for("carts.sounds_list", cart_id=cart_id)
    else:
        redirect_to = url_for("carts.index")

    return (
        "",
        200,
        {
            "HX-Redirect": redirect_to,
        },
    )


@blueprint.delete("/carts/<int:cart_id>")
@permission_required("carts")
async def delete(cart_id: int):
    cart = await Cart.from_id(g.dbsession, cart_id)
    if not cart:
        raise NotFound(f"Cart {cart_id} not found")
    shutil.rmtree(cart.path)
    await g.dbsession.delete(cart)
    await g.dbsession.commit()
    return ""


@blueprint.get("/carts/<int:cart_id>/sounds")
@login_required
async def sounds_list(cart_id: int):
    cart = await Cart.from_id(
        g.dbsession, cart_id, load_sounds=True, load_uploaders=True
    )
    if not cart:
        raise NotFound(f"Cart {cart_id} not found")
    next_sound = cart.next_sound(for_display=True)
    next_id = next_sound.id if next_sound else None
    return await render_template("carts/sounds.jinja", cart=cart, next_id=next_id)


async def _render_next_tags(cart_id: int, cart: Cart = None):
    """
    Renders the `next_tag` macro for all sounds in the cart. This will update
    (out-of-band) all tags' in the page.

    If `cart` is provided, it will be used instead of fetching the cart from
    the database.
    """
    if not cart:
        cart = await Cart.from_id(g.dbsession, cart_id, load_sounds=True)
    next_sound = cart.next_sound(for_display=True)
    next_id = next_sound.id if next_sound else None
    result = []
    for sound in cart.sounds:
        result.append(await render_macro("next_tag", sound=sound, next_id=next_id))
    return "\n".join(result)


@blueprint.post("/carts/<int:cart_id>/sounds")
@permission_required("carts")
async def add_sounds(cart_id: int):
    """
    This can be called from the sounds list, but also from carts list, when
    the cart is empty.
    """
    from_list = request.args.get("from_list", default=False)
    cart = await Cart.from_id(g.dbsession, cart_id)
    if not cart:
        raise NotFound(f"Cart {cart_id} not found")
    form = await stream_form(cart.path)
    added_sounds = []
    rank = await Sound.next_rank(g.dbsession, cart_id)
    for uploaded_file in form.getall("sounds"):
        metadata = mutagen.File(uploaded_file["uploaded_to"])
        if metadata is None:
            raise BadRequest(
                f"{uploaded_file['filename']} does not seem to be an audio file.",
            )
        sound = Sound(
            cart_id=cart_id,
            rank=rank,
            title=uploaded_file["filename"],
            path=uploaded_file["uploaded_to"],
            duration=int(metadata.info.length),
            uploader_id=current_user.user.id,
        )
        added_sounds.append(sound)
        g.dbsession.add(sound)
        rank += 1
    await g.dbsession.commit()

    await Scheduler.get().queue_analysis([s.id for s in added_sounds])

    g.dbsession.expire(cart)
    cart = await Cart.from_id(g.dbsession, cart_id, load_sounds=True)
    cart.recompute_duration()
    await g.dbsession.commit()
    if from_list:
        next_sound = cart.next_sound(for_display=True)
        next_id = next_sound.id if next_sound else None
        result = await render_macro(
            "sound_lines",
            sounds=added_sounds,
            next_id=next_id,
            mode=cart.mode,
            editable=current_user.user.can_carts(),
        )
        result += await _render_next_tags(None, cart)
        result += await render_macro("cart_summary", cart=cart)
        return result
    else:
        return (
            "",
            200,
            {
                "HX-Redirect": url_for("carts.sounds_list", cart_id=cart_id),
            },
        )


@blueprint.delete("/carts/<int:cart_id>/sounds/<int:sound_id>")
@permission_required("carts")
async def delete_sound(cart_id: int, sound_id: int):
    sound = await Sound.from_id(g.dbsession, sound_id)
    if not sound:
        raise NotFound(f"Sound {sound_id} not found")
    if not sound.cart.id == cart_id:
        raise NotFound(f"Sound {sound_id} is not in cart {cart_id}")

    sound.path.unlink()
    await g.dbsession.delete(sound)
    await g.dbsession.flush()

    cart = await Cart.from_id(g.dbsession, cart_id, load_sounds=True)
    result = await _render_next_tags(cart_id, cart=cart)
    result += await render_macro("cart_summary", cart=cart)

    sounds = await Sound.update_ranks(g.dbsession, cart_id)
    for sound in sounds:
        result += await render_macro("sound_rank", sound=sound)

    await g.dbsession.commit()
    return result


@blueprint.get("/carts/<int:cart_id>/sounds/<int:sound_id>")
@login_required
async def download_sound(cart_id: int, sound_id: int):
    sound = await Sound.from_id(g.dbsession, sound_id)
    if not sound:
        raise NotFound(f"Sound {sound_id} not found")
    if not sound.cart.id == cart_id:
        raise NotFound(f"Sound {sound_id} is not in cart {cart_id}")

    return await send_file(
        sound.path,
        mimetype="audio/mpeg",
        conditional=True,
        as_attachment=True,
        attachment_filename=sound.path.name,
    )


@blueprint.put("/carts/<int:cart_id>/sounds/<int:sound_id>")
@permission_required("carts")
async def edit_sound(cart_id: int, sound_id: int):
    sound = await Sound.from_id(g.dbsession, sound_id)
    if not sound:
        raise NotFound(f"Sound {sound_id} not found")
    if not sound.cart.id == cart_id:
        raise NotFound(f"Sound {sound_id} is not in cart {cart_id}")

    form = await request.form

    title = form.get("title")
    if title is not None:
        title = title.strip()
        if title:
            sound.title = title
        else:
            raise BadRequest("Please provide a title")

    active = form.get("active")
    if active is not None:
        sound.active = active == "true"

    await g.dbsession.commit()
    cart = await Cart.from_id(g.dbsession, cart_id, load_sounds=True)
    result = ""
    if title is not None:
        next_sound = cart.next_sound(for_display=True)
        next_id = next_sound.id if next_sound else None
        result += await render_macro(
            "sound_lines",
            sounds=[sound],
            next_id=next_id,
            mode=cart.mode,
            editable=current_user.user.can_carts(),
        )
    if active is not None:
        result += await render_macro("cart_summary", cart=cart)
        result += await _render_next_tags(None, cart=cart)

    return result, 200


@blueprint.post("/carts/<int:cart_id>")
@permission_required("carts")
async def sort_sounds(cart_id: int):
    result = ""

    form = await request.form
    required_ranks = form.getlist("rank")
    rank = 1
    for sound_id in required_ranks:
        sound = await Sound.from_id(g.dbsession, int(sound_id))
        if not sound:
            raise NotFound(f"Sound {sound_id} not found")
        if not sound.cart.id == cart_id:
            raise BadRequest(f"Sound {sound_id} is not in cart {cart_id}")
        sound.rank = rank
        result += await render_macro("sound_rank", sound=sound)
        rank += 1
    await g.dbsession.commit()

    result += await _render_next_tags(cart_id)

    return result, 200



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Radiotomate CLI
===============

All commands are implemented with Click, grouped as ``@radiotomate_cli.command()``, each
command in a sub-module. Do not forget to import the submodule in ``main``.

The radiotomate_cli group parses the configuration and configures loggers.
"""

from collections.abc import Mapping
from pathlib import Path

import click


@click.group()
@click.option(
    "-c",
    "--config-path",
    envvar="RTCONFIG",
    type=click.Path(
        readable=True,
        allow_dash=False,
        exists=True,
        resolve_path=True,
        path_type=Path,
    ),
    help=(
        "Path to the configuration file. "
        "You can also export this path in the RTCONFIG environment variable "
        "(this is recommended for regular use)."
    ),
)
@click.option("--verbose", is_flag=True, help="Sets logging level to DEBUG.")
@click.version_option()
@click.pass_context
def radiotomate_cli(ctx, verbose, config_path=None):
    main_handler = None
    if ctx.invoked_subcommand in ("interface", "scheduler"):
        main_handler = ctx.invoked_subcommand
    ctx.obj = LazyConfigLoader(config_path, verbose, main_handler)


class LazyConfigLoader(Mapping):
    """
    This should be used as Click's context object, to lazyly call `load_config`: it
    tries to load the config only when the command actually tries to read the dict.

    This avoids raising an error saying config path is mandatory in commands that don't
    need it (like install or --help)
    """

    def __init__(self, config_path, verbose, main_handler):
        self.config: dict = None
        self._config_path = config_path
        self._verbose = verbose
        self._main_handler = main_handler

    def _ensure_loaded(self):
        if not self.config:
            if not self._config_path:
                raise click.UsageError(
                    "Please set the RTCONFIG environment variable, or use:\n"
                    "    radiotomate -c CONFIG_PATH subcommand..."
                )
            self.config = load_config(
                self._config_path, self._verbose, self._main_handler
            )

    def __getitem__(self, key):
        self._ensure_loaded()
        return self.config[key]

    def __iter__(self):
        self._ensure_loaded()
        return iter(self.config)

    def __len__(self):
        self._ensure_loaded()
        return len(self.config)


def load_config(config_path, verbose=False, main_handler=None) -> dict:
    import logging.config

    from ruamel.yaml import YAML

    yaml = YAML(typ="safe")
    with config_path.open() as f:
        config = yaml.load(f)

    if verbose:
        config["logging"]["loggers"]["radiotomate"]["level"] = "DEBUG"

    if not main_handler:
        main_handler = "default"
    config["logging"]["root"]["handlers"].append(main_handler)
    config["logging"]["loggers"]["radiotomate"]["handlers"].append(main_handler)

    logging.config.dictConfig(config["logging"])

    return config



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from __future__ import annotations

import logging
from datetime import datetime
from pathlib import Path  # noqa: TC003
from random import choice

from apscheduler.triggers.cron import CronTrigger
from sqlalchemy import Boolean, Integer, Select, String, case, delete, func, select
from sqlalchemy.dialects.sqlite import DATETIME
from sqlalchemy.orm import (
    Mapped,
    Session,
    mapped_column,
    query_expression,
    relationship,
    selectinload,
    with_expression,
)

from radiotomate.db import Base
from radiotomate.enums import CartMode, ScheduleMode
from radiotomate.models.sound import Sound

_log = logging.getLogger(__name__)

# Sound carts having this value as Cart.url should be pushed to the auto-dj queue
# instead of the usual carts queue
URL_TO_AUTODJ_QUEUE = "autodj"


class Cart(Base):
    """
    Sound cart: contains sounds that can be scheduled to play.

    This class does **not** manage the cart's folder, it's only stored in
    ``path`` for future reference.

    Note that path is relative to the app's data root, but absolute when instanciated
    in Python. This is provided by `radiotomate.db`.
    """

    __tablename__ = "carts"
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    title: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    path: Mapped[Path]  # should be null only while creating the cart's folder
    url: Mapped[str]
    average_duration: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    max_duration: Mapped[int] = mapped_column(Integer)
    notes: Mapped[str] = mapped_column(String, nullable=False, default="")
    created: Mapped[datetime] = mapped_column(DATETIME, default=datetime.now)
    modified: Mapped[datetime] = mapped_column(
        DATETIME,
        default=datetime.now,
        onupdate=datetime.now,
    )

    mode: Mapped[CartMode]
    schedule_mode: Mapped[ScheduleMode]
    schedule_correct: Mapped[bool] = mapped_column(
        Boolean,
        nullable=False,
        default=False,
    )

    #: all schedule filter columns default to * to ease tests
    schedule_year: Mapped[str] = mapped_column(String, nullable=False, default="*")
    schedule_month: Mapped[str] = mapped_column(String, nullable=False, default="*")
    schedule_day: Mapped[str] = mapped_column(String, nullable=False, default="*")
    schedule_week: Mapped[str] = mapped_column(String, nullable=False, default="*")

    #: as in APScheduler, 0 is monday
    schedule_day_of_week: Mapped[str] = mapped_column(
        String,
        nullable=False,
        default="*",
    )
    schedule_hour: Mapped[str] = mapped_column(String, nullable=False, default="*")
    schedule_minute: Mapped[str] = mapped_column(String, nullable=False, default="*")
    schedule_second: Mapped[str] = mapped_column(String, nullable=False, default="*")

    sounds: Mapped[list[Sound]] = relationship(
        lazy=True,
        back_populates="cart",
        order_by="Sound.rank",
        cascade="save-update, merge, expunge, delete",
    )
    sounds_count: Mapped[int] = query_expression()
    inactive_sounds_count: Mapped[int] = query_expression()

    @classmethod
    def add_query_expressions(cls, query: Select) -> Select:
        """
        This is an alternative to `q.options(selectinload(Cart.sounds))`: it adds
        mapped query_expression to a query selecting from Cart, including the outer
        join and group by cart.
        """
        return (
            query.join_from(Cart, Cart.sounds, isouter=True)
            .group_by(Cart.id)
            .options(
                with_expression(Cart.sounds_count, func.count(Cart.sounds)),
                with_expression(
                    Cart.inactive_sounds_count,
                    func.sum(
                        case(
                            (Sound.active.is_(False), 1),
                            (Sound.gain.is_(None), 1),
                            else_=0,
                        )
                    ),
                ),
            )
        )

    def postfill_query_expressions(self) -> None:
        """
        Compute query-expression fields when Cart.sounds is fully loaded
        """
        self.sounds_count = len(self.sounds)

        def inactive(sound: Sound) -> int:
            if not sound.available:
                return 1
            else:
                return 0

        self.inactive_sounds_count = sum(map(inactive, self.sounds))

    @classmethod
    async def all(
        cls,
        session: Session,
        schedule_mode: ScheduleMode | None = None,
        load_sounds: bool = False,
    ) -> list[Cart]:
        """
        Returns all carts, with sounds_count pre-computed.

        If ``schedule_mode`` is provided, will only select carts having the given value.

        Associated sounds are join-loaded if ``load_sounds``.
        """
        q = select(Cart)
        if schedule_mode:
            q = q.filter(Cart.schedule_mode == schedule_mode)
        if load_sounds:
            q = q.options(selectinload(Cart.sounds))
        else:
            q = cls.add_query_expressions(q)
        carts = (await session.scalars(q)).all()
        if load_sounds:
            for cart in carts:
                cart.postfill_query_expressions()
        return carts

    @classmethod
    async def from_title(cls, session: Session, title: str) -> Cart | None:
        return await session.scalar(select(Cart).filter(Cart.title == title))

    @classmethod
    async def from_id(
        cls,
        session: Session,
        cart_id: int,
        load_sounds: bool = False,
        load_uploaders: bool = False,
    ) -> Cart | None:
        """
        Returns a cart. `sounds_count` is pre-computed, or associated sounds
        are preloaded if `load_sounds` is True.
        `load_uploaders` requires `load_sounds` and will also pre-load
        `cart.sound[i].uploader`.
        """
        q = select(Cart)
        if load_sounds:
            if load_uploaders:
                q = q.options(selectinload(Cart.sounds).selectinload(Sound.uploader))
            else:
                q = q.options(selectinload(Cart.sounds))
        else:
            q = cls.add_query_expressions(q)
        q = q.filter(Cart.id == cart_id)
        cart = await session.scalar(q)
        if load_sounds:
            cart.postfill_query_expressions()
        return cart

    @classmethod
    async def delete(cls, session: Session, cart_id: int):
        await session.execute(delete(Cart).where(Cart.id == cart_id))

    @property
    def schedule_repr(self) -> str:
        return "[{},{}] at {} {} {} {} {} {} {} {}".format(
            self.schedule_mode,
            "OK" if self.schedule_correct else "error",
            self.schedule_year,
            self.schedule_month,
            self.schedule_day,
            self.schedule_week,
            self.schedule_day_of_week,
            self.schedule_hour,
            self.schedule_minute,
            self.schedule_second,
        )

    @property
    def schedule_is_advanced(self) -> bool:
        if (
            self.schedule_mode is ScheduleMode.TIMED
            and self.schedule_year == "*"
            and self.schedule_month == "*"
            and self.schedule_day == "*"
            and self.schedule_week == "*"
            and self.schedule_second == "0"
        ):
            try:
                _ = int(self.schedule_day_of_week)
                _ = int(self.schedule_hour)
                _ = int(self.schedule_minute)
                return False
            except ValueError:
                return True
        return self.schedule_mode is ScheduleMode.TIMED

    def to_timed_trigger(self) -> CronTrigger:
        """
        Returns a CronTrigger object from a cart's schedule columns, when it's `timed`.
        Raises an IndexError if the schedule is invalid.

        Note: APScheduler's week days are 0-7 for sun,mon,tue,wed,thu,fri,sat, sun
        """
        return CronTrigger(
            year=self.schedule_year,
            month=self.schedule_month,
            day=self.schedule_day,
            week=self.schedule_week,
            day_of_week=self.schedule_day_of_week,
            hour=self.schedule_hour,
            minute=self.schedule_minute,
            second=self.schedule_second,
        )

    def recompute_duration(self):
        self.average_duration = sum([s.duration for s in self.sounds]) // len(
            self.sounds,
        )

    def next_sound(self, for_display=False) -> Sound | None:
        """
        ``for_display`` tells if the method is called to display the next sound
        or (by default) if it's called to actually play the sound. This avoids
        displaying a false information for carts playing randomly.
        """
        if for_display and self.mode is CartMode.RANDOM:
            return None
        implem = getattr(self, "next_sound_" + self.mode.value)
        return implem()

    def next_sound_playlist(self) -> Sound | None:
        for sound in self.sounds:
            if sound.available and not sound.last_played:
                return sound

    def next_sound_playlist_loop(self) -> Sound | None:
        last_played = datetime(2000, 1, 1)
        last_played_rank = None
        for sound in self.sounds:
            if sound.last_played and sound.last_played > last_played:
                last_played_rank = sound.rank
                last_played = sound.last_played

        if last_played_rank:
            for sound in self.sounds:
                if sound.available and sound.rank > last_played_rank:
                    return sound

        for sound in self.sounds:
            if sound.available:
                return sound

        return None

    def next_sound_random(self) -> Sound | None:
        available_sounds = [s for s in self.sounds if s.available]
        if available_sounds:
            return choice(available_sounds)
        else:
            return None



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Scheduling-related endpoints and tasks.

Note that in this module Cart.id is usually a string, because it is also used as a
Schedule ID and APScheduler prefers strings.
"""

import json
import logging
from datetime import datetime, timedelta

from apscheduler import CoalescePolicy, ScheduleLookupError
from apscheduler.triggers.date import DateTrigger
from httpx import AsyncClient
from quart import Blueprint, current_app, g
from sqlalchemy.orm import Session as ormSession
from werkzeug.exceptions import BadRequest

from radiotomate.auth import token_required
from radiotomate.enums import CartMode, ScheduleMode
from radiotomate.models import Cart
from radiotomate.models.cart import URL_TO_AUTODJ_QUEUE

_log = logging.getLogger(__name__)

blueprint = Blueprint("schedule", __name__)


async def skip_cart(session: ormSession, current_app, sound_id: int | None = None):
    _log.debug("starting skip_cart(%r)", sound_id)
    playout_client: AsyncClient = current_app.config["PLAYOUT_CLIENT"]
    if sound_id is None:
        params = {"relay": 1}
    else:
        params = {"radiotomate_sound_id": sound_id}
    response = await playout_client.delete("/live", params=params)
    if response.status_code == 204:
        _log.info("No max duration to enforce for sound#%r", sound_id)
    elif response.status_code != 200:
        msg = response.text
        _log.error("Enforcing max duration on sound#%r failed: %s", sound_id, msg)


async def push_cart(session: ormSession, current_app, cart_id: int):
    _log.debug("starting push_cart(%d)", cart_id)
    playout_client = current_app.config["PLAYOUT_CLIENT"]
    cart = await Cart.from_id(session, cart_id, load_sounds=True)
    if not cart:
        _log.warning(f"Scheduled cart #{cart_id} not found")
        return
    if cart.mode == CartMode.RELAY:
        await _push_cart_relay(cart, playout_client, current_app)
    else:
        await _push_cart_sound(cart, playout_client, current_app)


async def _push_cart_sound(cart: Cart, playout_client: AsyncClient, current_app):
    sound = cart.next_sound()
    if sound:
        queue = "carts"
        if cart.url == URL_TO_AUTODJ_QUEUE:
            queue = URL_TO_AUTODJ_QUEUE
        response = await playout_client.post(
            f"/queue/{queue}",
            json={
                "path": str(sound.path),
                "artist": cart.title,
                "title": sound.title,
                "radiotomate_sound_id": sound.id,
                "rg_track_gain": str(sound.gain),
            },
        )
        if response.status_code == 200:
            result = response.json()
            _log.info(
                "Pushed cart %d sound %s:%s as RID %s",
                cart.id,
                sound.id,
                sound.path,
                result,
            )
            if cart.max_duration:
                # remove a potential previous entry
                await current_app.scheduler.remove_schedule(str(-cart.id))
                when = datetime.now() + timedelta(seconds=cart.max_duration)
                await current_app.scheduler.add_schedule(
                    skip_cart,
                    DateTrigger(when),
                    id=str(-cart.id),
                    args=[int(sound.id)],
                    coalesce=CoalescePolicy.latest,
                    misfire_grace_time=10.0,
                )

        else:
            try:
                result = response.json()
            except json.JSONDecodeError:
                result = response.text
            _log.error(
                "Error while pushing cart %d sound %s:%s: %s",
                cart.id,
                sound.id,
                sound.path,
                result,
            )
    else:
        _log.warning("Scheduled cart %d:%s has no next sound", cart.id, cart.title)


async def _push_cart_relay(cart: Cart, playout_client: AsyncClient, current_app):
    response = await playout_client.post("/relay", params={"url": cart.url})
    if response.status_code == 200:
        result = response.json()
        _log.info(
            "Pushed relay cart %d: %s | status: %s",
            cart.id,
            cart.url,
            result,
        )
        if cart.max_duration:
            # remove a potential previous entry
            await current_app.scheduler.remove_schedule(str(-cart.id))
            when = datetime.now() + timedelta(seconds=cart.max_duration)
            await current_app.scheduler.add_schedule(
                skip_cart,
                DateTrigger(when),
                id=str(-cart.id),
                coalesce=CoalescePolicy.latest,
                misfire_grace_time=10.0,
            )
        else:
            _log.warning(
                "relay cart %d has no maximum duration! might play forever.", cart.id
            )
    else:
        try:
            result = response.json()
        except json.JSONDecodeError:
            result = response.text
        _log.error("Error while pushing relay cart %d:%s: %s", cart.id, result)


@blueprint.put("/schedule/<cart_id>")
@token_required
async def post_schedule(cart_id: str):
    try:
        cart = await Cart.from_id(g.dbsession, int(cart_id))
    except ValueError:
        return f"cart #{cart_id} not found", 404
    if not cart:
        return f"cart #{cart_id} not found", 404

    await current_app.scheduler.remove_schedule(cart_id)

    try:
        if cart.schedule_mode is ScheduleMode.TIMED:
            trigger = cart.to_timed_trigger()
        else:
            message = f"Requested to schedule the non-timed cart #{cart_id}: "
            message += cart.schedule_repr
            _log.error(message)
            raise BadRequest(message)
    except ValueError:
        message = f"Invalid cron expression for cart #{cart_id}: {cart.schedule_repr}"
        _log.error(message)
        cart.schedule_correct = False
        await g.dbsession.commit()
        raise BadRequest(message) from None

    await current_app.scheduler.add_schedule(
        push_cart,
        trigger,
        id=cart_id,
        args=[int(cart_id)],
        coalesce=CoalescePolicy.latest,
        misfire_grace_time=10.0,
    )

    cart.schedule_correct = True
    return "", 200


@blueprint.get("/schedule/<cart_id>")
@token_required
async def get_schedule(cart_id: str):
    try:
        schedule = await current_app.scheduler.get_schedule(cart_id)
        return {
            "next_time": schedule.next_fire_time.isoformat(),
        }
    except ScheduleLookupError:
        return f"cart #{cart_id} not scheduled", 404



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
# Radiotomate's enumerated types

This contains fixed-list values that are not worth a DB table,
but we don't want to hardcode their string value either.

Those are in their own module to avoid dependency loops.
"""

from enum import Enum, IntEnum, StrEnum


class CartMode(Enum):
    """
    Cart.mode
    """

    PLAYLIST = "playlist"
    PLAYLIST_LOOP = "playlist_loop"
    RANDOM = "random"
    RELAY = "relay"

    @property
    def label(self):
        """
        TODO this should be replaced by translation functions
        """
        match self:
            case CartMode.PLAYLIST:
                return "Playlist"
            case CartMode.PLAYLIST_LOOP:
                return "Looping Playlist"
            case CartMode.RANDOM:
                return "Random"
            case CartMode.RELAY:
                return "Relay"


class ScheduleMode(Enum):
    JINGLES = "jingles"
    TIMED = "timed"

    @property
    def label(self):
        """
        TODO this should be replaced by translation functions
        """
        match self:
            case ScheduleMode.JINGLES:
                return "Jingles"
            case ScheduleMode.TIMED:
                return "Timed"


class Role(StrEnum):
    ADMIN = "admin"
    STREAM = "stream"
    LIVE = "live"
    CARTS = "carts"
    NOT_AUDIO = "not_audio"
    AUTODJ = "autodj"


class DayNames(IntEnum):
    MONDAY = 0
    TUESDAY = 1
    WEDNESDAY = 2
    THURSDAY = 3
    FRIDAY = 4
    SATURDAY = 5
    SUNDAY = 6

    @property
    def display(self):
        """
        TODO this should be replaced by translation functions
        """
        return self.name.capitalize()

    def matches(self, other: int | str) -> bool:
        if isinstance(other, str):
            return str(self.value) == other
        else:
            return self.value == other



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import logging
from importlib.metadata import version

_log = logging.getLogger(__name__)


def get_version() -> str:
    """
    Get Radiotomate's version from package metadata
    """
    try:
        return version("radiotomate")
    except Exception as excn:
        _log.warning(excn)
    return "demo"



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from __future__ import annotations

import logging
from datetime import datetime

from sqlalchemy import ForeignKey, Integer, String, select
from sqlalchemy.dialects.sqlite import DATETIME, JSON
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.orm import Session as ormSession

from radiotomate.db import Base
from radiotomate.models.sound import Sound

_log = logging.getLogger(__name__)


class MetadataLog(Base):
    """
    Metadata saved from the playout process: this is what actually got broadcasted.
    """

    __tablename__ = "metadata_log"
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    on_air: Mapped[datetime] = mapped_column(DATETIME, default=datetime.now, index=True)
    source: Mapped[str] = mapped_column(String)
    source_url: Mapped[str] = mapped_column(String)
    cart_id: Mapped[int] = mapped_column(Integer, ForeignKey("carts.id"))
    artist: Mapped[str] = mapped_column(String)
    title: Mapped[str] = mapped_column(String)
    album: Mapped[str] = mapped_column(String)
    extra: Mapped[dict] = mapped_column(
        MutableDict.as_mutable(JSON),
        default={},
        nullable=False,
    )  # https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#sqlalchemy.dialects.sqlite.JSON

    cart: Mapped[Cart] = relationship()  # noqa: F821

    @classmethod
    async def from_playout(cls, session: ormSession, md: dict) -> MetadataLog:  # noqa: PLR0912
        """
        Transform the JSON metadata from the playout process into a MetadataLog object.
        """
        parsed = MetadataLog()
        if "on_air" in md:
            if md["on_air"]:
                parsed.on_air = datetime.fromisoformat(md["on_air"])
            del md["on_air"]
        if "radiotomate_sound_id" in md:
            if md["radiotomate_sound_id"]:
                sound = await Sound.from_id(session, md["radiotomate_sound_id"])
                if sound:
                    sound.last_played = parsed.on_air
                    parsed.cart = sound.cart
            del md["radiotomate_sound_id"]
        if "source" in md:
            if md["source"]:
                parsed.source = md["source"]
            del md["source"]
        if "source_url" in md:
            if md["source_url"]:
                parsed.source_url = md["source_url"]
            del md["source_url"]
        if "initial_uri" in md:
            if md["initial_uri"]:
                # we push annotated requests to liquidsoap, but we don't want to store
                # annotations
                raw_uri = md["initial_uri"]
                if raw_uri.startswith("annotate:"):
                    closing_colon = raw_uri.find(":", 9)
                    if closing_colon > 0:
                        raw_uri = raw_uri[closing_colon + 1 :]
                parsed.source_url = raw_uri
            del md["initial_uri"]
        if "artist" in md:
            if md["artist"]:
                parsed.artist = md["artist"]
            del md["artist"]
        if "title" in md:
            if md["title"]:
                parsed.title = md["title"]
            del md["title"]
        if "album" in md:
            if md["album"]:
                parsed.album = md["album"]
            del md["album"]
        parsed.extra = md
        return parsed

    @classmethod
    async def get(
        cls,
        session: ormSession,
        limit: int | None = None,
    ) -> list[MetadataLog]:
        """
        Returns a list of MetadataLog objects, ordered by `on_air` DESC.
        """
        q = select(MetadataLog)
        if limit is not None:
            q = q.limit(limit)
        q = q.order_by(MetadataLog.on_air.desc())
        return list(await session.scalars(q))

    @property
    def source_filename(self) -> str:
        """
        Returns the last part of self.source_url, if possible
        """
        if self.source_url:
            url = str(self.source_url)
            last_slash = url.rfind("/")
            if last_slash >= 0:
                return url[last_slash + 1 :]
            return url
        return ""



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
This module defines endpoint that trigger sounds' replaygain analysis.

It relies on beets' replaygain plug-in, with a nicer thread pool defined in
`BeetsIntegration`.
"""

import logging

from quart import Blueprint, request
from werkzeug.exceptions import BadRequest

from radiotomate.auth import token_required
from radiotomate.beets import BeetsIntegration

_log = logging.getLogger(__name__)

blueprint = Blueprint("analyzer", __name__)


@blueprint.post("/analyzer")
@token_required
async def analyzer_post():
    """
    Start a replay gain analysis on given Sound.
    """
    if not request.is_json:
        raise BadRequest("JSON object expected")
    data = await request.get_json()
    beets = BeetsIntegration.get()
    try:
        for sound_id in data["sound_ids"]:
            await beets.analyze_soon(sound_id)
    except (KeyError, ValueError) as e:
        raise BadRequest(
            "We expect a JSON object with a single field 'sound_ids',an interger list"
        ) from e
    return {}



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
This module contains classes and wrappers to interface more easily with Beets.
"""

from __future__ import annotations

import logging
from asyncio import CancelledError, Queue, get_running_loop, sleep
from optparse import Values
from random import randint
from typing import TYPE_CHECKING

from beets.dbcore import sort_from_strings
from beets.library import Item, Library
from beets.ui import _setup
from beetsplug.replaygain import ReplayGainPlugin
from quart import Quart, current_app

from radiotomate.models.sound import Sound

if TYPE_CHECKING:
    from radiotomate.db import QuartAlchemy

_log = logging.getLogger(__name__)


class SoundToItemProxy(Item):
    "Masquerades one of our Sound into a beet's Item"

    def __init__(self, sound: Sound):
        super().__init__()
        self._sound = sound
        self.path = str(sound.path.absolute()).encode()
        self.rg_track_gain = sound.gain
        self.rg_track_peak = sound.peak

    def store(self, fields=None):
        self._sound.gain = self.rg_track_gain
        self._sound.peak = self.rg_track_peak


class BeetsIntegration:
    """
    This class wraps a Beets library and provides high-level helpers for our AutoDJ and
    replaygain analysis (this implies that replaygain level is set in the Beets
    configuration).

    It can also register as a Quart extension, so its instance can be easily found by
    views and awaited.
    """

    EXTENSION_NAME = "RadiotomateBeetsIntegration"

    # background analysis are started only after this delay, so the application and
    # playout process have all CPU available while starting
    BACKGROUND_ANALYSIS_DELAY = 30.0

    def __init__(self, lib: Library):
        self.lib: Library = lib
        self.rg = ReplayGainPlugin()
        self.analyzer_q = Queue()
        # logging.getLogger("beets.replaygain").setLevel(logging.DEBUG)

    @classmethod
    def default(cls) -> BeetsIntegration:
        """
        Instantiates Beets using its default config path
        """
        options = Values(
            {
                "config": None,  # could be a config path
                "exclude": None,
                "plugins": None,
            }
        )
        _, lib = _setup(options, None)
        return BeetsIntegration(lib)

    @classmethod
    def get(cls) -> BeetsIntegration:
        """
        Find the instance wrapped in ``current_app``
        """
        return current_app.extensions[cls.EXTENSION_NAME]

    def init_app(self, app: Quart) -> None:
        if self.EXTENSION_NAME in app.extensions:
            raise RuntimeError(
                "An sqlalchemy extension has already been registered on this app",
            )
        app.extensions[self.EXTENSION_NAME] = self

    async def random_pick(self, query) -> Item | None:
        """
        Like "beet random [query]", returns the path ot a track matching given query
        """

        def fetch_items(q):
            return list(self.lib.items(q))

        loop = get_running_loop()
        objs = await loop.run_in_executor(None, fetch_items, query)
        if objs:
            index = randint(0, len(objs) - 1)
            obj = objs[index]
            return obj
        else:
            return None

    async def search(self, query) -> list[Item]:
        """
        Like "beet [query]", returns matching Items
        """

        def fetch_items(q):
            sort = sort_from_strings(Item, ["artist+"])
            return list(self.lib.items(q, sort=sort))

        loop = get_running_loop()
        objs = await loop.run_in_executor(None, fetch_items, query)
        return objs

    async def background_analyzer(self, db: QuartAlchemy):
        """
        This should be started as a background task
        """
        await sleep(self.BACKGROUND_ANALYSIS_DELAY)
        async with db.session() as session:
            ids = await Sound.ids_without_gain(session)
            for i in ids:
                await self.analyzer_q.put(i)
            if ids:
                _log.info(
                    "Queue'd %d sounds catching up their replaygain analysis", len(ids)
                )

        sound_id = None
        while True:
            try:
                sound_id = await self.analyzer_q.get()
                async with db.session() as dbsession:
                    sound = await Sound.from_id(dbsession, sound_id)
                    if sound:
                        await self._do_analyze_rg(sound)
                        await dbsession.commit()
            except CancelledError:
                return
            except Exception:
                _log.exception("while attempting sound_id=%r", sound_id)

    async def _do_analyze_rg(self, sound: Sound):
        """
        Fill ReplayGain columns on sound
        """
        proxied = SoundToItemProxy(sound)
        loop = get_running_loop()
        await loop.run_in_executor(None, self.rg.handle_track, proxied, False)

    async def analyze_soon(self, sound_id: int):
        """
        Add a sound to the replaygain analysis queue
        """
        await self.analyzer_q.put(sound_id)



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
# Root module of the Interface Web application

All endpoints are defined in sub-modules, with one blueprint per sub-module.

This module contains helper functions.
"""

from collections.abc import Coroutine
from contextlib import ExitStack
from pathlib import Path
from time import time_ns

import quart
from multipart import (
    MultiDict,
    MultipartSegment,
    PushMultipartParser,
    parse_options_header,
)

_allowed_chars = ("-", ".", "_")


def safe_path(basedir: Path, filename: str, prefix: str = "") -> Path:
    """
    Generates a complete path in `basedir` from `filename`,
    and ensures nothing already exists there.
    Keeps only alphanumeric characters and the characters in _allowed_chars,
    also prefix the files with `prefix`, or 4 random digits to avoid collisions.
    """
    if not basedir:
        raise ValueError("Missing basedir")
    safe_filename = "".join(c for c in filename if c.isalnum() or c in _allowed_chars)
    while True:
        target_path = basedir / (prefix + safe_filename)
        if not target_path.exists():
            return target_path
        prefix = str(time_ns())[-4:]


def safe_int(s: str | None, default: int | None, positive=False) -> int | None:
    """
    Parses given string (extracted from a form dict) to an integer. Returns `default`
    if the value cannot be parsed to an integer. When `positive=True`, returned
    `default` if parsed value is negative.
    """
    if s is None:
        return default
    try:
        parsed = int(s)
        if positive and parsed < 0:
            return default
        else:
            return parsed
    except ValueError:
        return default


async def stream_form(
    base_path: Path | None = None,
) -> Coroutine[None, None, MultiDict]:
    header = quart.request.headers.get("Content-Type")
    _, options = parse_options_header(header)
    boundary = options["boundary"]
    form_fields = MultiDict()
    current_form_field = None
    current_file = None
    with ExitStack() as stack:
        parser = PushMultipartParser(boundary, strict=True)
        stack.enter_context(parser)

        async for chunk in quart.request.body:
            for result in parser.parse(chunk):
                if isinstance(result, MultipartSegment):
                    if result.filename:
                        target_path = safe_path(base_path, result.filename)
                        current_file = target_path.open("wb")
                        stack.enter_context(current_file)
                        form_fields[result.name] = {
                            "filename": result.filename,
                            "content_type": result.content_type,
                            "uploaded_to": target_path,
                        }
                    else:
                        current_form_field = result.name
                elif result:  # Result is a non-empty bytearray
                    if current_form_field:
                        form_fields[current_form_field] = result.decode()
                    else:
                        current_file.write(result)
                else:  # end of segment
                    current_form_field = None
                    current_file = None
            if parser.closed:
                break
    return form_fields



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import logging

from quart import Blueprint, send_file
from werkzeug.exceptions import NotFound

from radiotomate.auth import login_required
from radiotomate.beets import BeetsIntegration

_log = logging.getLogger(__name__)

blueprint = Blueprint("library", __name__, template_folder="templates")


@blueprint.get("/library/download/<int:item_id>")
@login_required
async def download(item_id: int):
    beets = BeetsIntegration.get()
    item = beets.lib.get_item(item_id)
    if item is None:
        raise NotFound()
    return await send_file(
        item.filepath,
        mimetype="audio/mpeg",
        conditional=True,
        as_attachment=True,
        attachment_filename=item.filepath.name,
    )



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import logging
from random import randint, sample

from quart import Blueprint, g, render_template, request, url_for
from werkzeug.datastructures import MultiDict
from werkzeug.exceptions import BadRequest, Conflict, NotFound

from radiotomate.auth import login_required, permission_required
from radiotomate.beets import BeetsIntegration
from radiotomate.enums import DayNames
from radiotomate.interface import safe_int
from radiotomate.models import AutoDJSlot

_log = logging.getLogger(__name__)

blueprint = Blueprint("autodj", __name__, template_folder="templates")

REM_PER_HOUR = 4
MIN_PER_DAY = 24 * 60


def random_color():
    """
    Create a mid-saturated color
    """
    r = randint(100, 200)
    g = randint(100, 200)
    b = randint(100, 200)
    return f"#{r:02x}{g:02x}{b:02x}"


def process_filters(form: MultiDict, slot: AutoDJSlot):
    """
    Transform POSTed filters into a ``slot.constraints`` dict.

    This dict will contain the total of filters' weights, and a list of tuples: first
    item is the filter's weight, second  is the filter formular. Thus a weighted
    selector only has to pick an integer in [0, max], then iterate in the
    list to find the first couple having a cumulated weight greater than picked number.
    """
    weights = form.getlist("filters[][weight]")
    filters = form.getlist("filters[][filter]")
    totalweight = 0
    constraints = {
        "filters": [],
    }
    for i in range(len(weights)):
        query = filters[i].strip()
        weight = max(1, safe_int(weights[i], 1, True))
        totalweight += weight
        constraints["filters"].append([weight, query])
    constraints["totalweight"] = totalweight
    slot.constraints = constraints


@blueprint.get("/autodj")
@login_required
async def index():
    slots_per_day = {}
    for d in range(7):
        slots_per_day[d] = []
    for slot in await AutoDJSlot.all(g.dbsession):
        slots_per_day[slot.day_of_week].append(slot)
    for d in range(7):
        day = slots_per_day[d]
        day.append(None)  # because the loop updates "previous" - should be popped after
        previous = None

        # we want each entry to measure at least 1 REM high. In order to keep next ones
        # as much aligned as possible, we cumulate the created overflow, to maybe make
        # following longer slots thinner.
        overflow = 0
        for slot in day:
            if previous:
                if slot:
                    ending = slot.minute
                else:
                    ending = MIN_PER_DAY
                h = REM_PER_HOUR * ((ending - previous.minute) / 60)
                if h < 1:
                    overflow += 1 - h
                    h = 1
                else:
                    margin = h - 1
                    if overflow <= margin:
                        h -= overflow
                        overflow = 0
                    else:
                        h = 1
                        overflow -= margin

                previous.height = h

            previous = slot
        day.pop()  # remove the fake entry

    return await render_template(
        "autodj/index.jinja",
        slots_per_day=slots_per_day,
        day_names=DayNames,
        REM_PER_HOUR=REM_PER_HOUR,
    )


@blueprint.get("/autodj/add")
@permission_required("autodj")
async def add_form():
    return await render_template(
        "autodj/slot_form.jinja", color=random_color(), day_names=DayNames
    )


@blueprint.post("/autodj")
@permission_required("autodj")
async def add():
    form = await request.form
    title = form.get("title")
    color = form.get("color")
    if len(color) != 7:
        color = random_color()
    hour = min(23, safe_int(form.get("hour"), 0, True))
    minute = 60 * hour + min(59, safe_int(form.get("minute"), 0, True))

    day_option = form.get("day_of_week")
    existing = None
    if not day_option:
        raise BadRequest("Please choose a day of week")
    elif day_option == "*":
        existing = await AutoDJSlot.from_time(g.dbsession, minute, exact=True)
        days_of_week = range(7)
    else:
        day_of_week = min(6, safe_int(day_option, 0, True))
        existing = await AutoDJSlot.from_time(
            g.dbsession, minute, day_of_week=day_of_week, exact=True
        )
        days_of_week = [day_of_week]

    if existing:
        if existing.title:
            hint = f"({existing.title}) "
        else:
            hint = ""
        raise Conflict(f"A slot {hint}is already scheduled at that time.")

    for d in days_of_week:
        slot = AutoDJSlot(title=title, color=color, day_of_week=d, minute=minute)
        process_filters(form, slot)
        g.dbsession.add(slot)

    await g.dbsession.commit()
    return (
        "",
        200,
        {
            "HX-Redirect": url_for("autodj.index"),
        },
    )


@blueprint.get("/autodj/<int:slot_id>")
@permission_required("autodj")
async def get(slot_id: int):
    slot = await AutoDJSlot.from_id(g.dbsession, slot_id)
    if not slot:
        raise NotFound("Auto-DJ slot not found")
    return await render_template(
        "autodj/slot_form.jinja", slot=slot, color=slot.color, day_names=DayNames
    )


@blueprint.put("/autodj/<int:slot_id>")
@permission_required("autodj")
async def edit(slot_id: int):
    slot = await AutoDJSlot.from_id(g.dbsession, slot_id)
    if not slot:
        raise NotFound("Auto-DJ slot not found")

    form = await request.form
    slot.title = form.get("title")
    color = form.get("color")
    if len(color) == 7:
        slot.color = color
    hour = min(23, safe_int(form.get("hour"), 0, True))
    minute = 60 * hour + min(59, safe_int(form.get("minute"), 0, True))
    day_of_week = min(6, safe_int(form.get("day_of_week"), 0, True))

    existing = await AutoDJSlot.from_time(
        g.dbsession, minute, day_of_week=day_of_week, exact=True
    )
    if existing and existing.id != slot.id:
        if existing.title:
            hint = f"({existing.title}) "
        else:
            hint = ""
        raise Conflict(f"A slot {hint}is already scheduled at that time.")

    slot.minute = minute
    slot.day_of_week = day_of_week

    process_filters(form, slot)

    await g.dbsession.commit()
    return (
        "",
        200,
        {
            "HX-Redirect": url_for("autodj.index"),
        },
    )


@blueprint.delete("/autodj/<int:slot_id>")
@permission_required("autodj")
async def delete(slot_id: int):
    slot = await AutoDJSlot.from_id(g.dbsession, slot_id)
    if not slot:
        raise NotFound("Auto-DJ slot not found")
    await g.dbsession.delete(slot)
    await g.dbsession.commit()

    return (
        "",
        200,
        {
            "HX-Redirect": url_for("autodj.index"),
        },
    )


@blueprint.get("/autodj/testquery")
@login_required
async def testquery():
    query = request.args.get("filters[][filter]")
    beets = BeetsIntegration.get()
    results = await beets.search(query)
    nb_results = len(results)
    if nb_results > 10:
        results = sample(results, k=10)
    return await render_template(
        "autodj/testquery.jinja",
        query=query,
        results=results,
        nb_results=nb_results,
    )



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import logging

from quart import Blueprint, render_template

from radiotomate.auth import login_required

_log = logging.getLogger(__name__)

blueprint = Blueprint("home", __name__, template_folder="templates")
watching_clients = set()


@blueprint.get("/")
@login_required
async def index():
    return await render_template("home/index.jinja")



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
# Root module of the Scheduler Web application

All endpoints are defined in sub-modules, with one blueprint per sub-module.
"""



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import click

from . import radiotomate_cli


@radiotomate_cli.command()
@click.option(
    "--reload",
    is_flag=True,
    help="Reloads app on changes (when developing)",
)
@click.pass_obj
def scheduler(config_dict, reload):
    """
    Start the Radiotomate scheduler service
    """
    import asyncio

    from hypercorn.asyncio import serve

    from radiotomate.beets import BeetsIntegration
    from radiotomate.quart import CustomQuart, ShutdownManager
    from radiotomate.scheduler.watchdog import Watchdog
    from radiotomate.scheduler_app import app_factory

    beets = BeetsIntegration.default()
    app = app_factory(config_dict, beets)

    Watchdog(app)

    if reload:
        app.run(use_reloader=reload, host="0.0.0.0", port=6822)
    else:

        @app.before_serving
        async def on_start():
            ShutdownManager().setup_signal_handler()

        config = CustomQuart.make_basic_hypercorn_config()
        config.bind = ["0.0.0.0:6822"]
        asyncio.run(serve(app, config, shutdown_trigger=ShutdownManager().wait))



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import asyncio
import logging

from quart import Quart

from radiotomate.quart import ShutdownError, or_shutdown

_log = logging.getLogger(__name__)


class Watchdog:
    """
    This registers as an application extension, to ensure only one instance will
    ever exists. It expects a completely configured scheduler app, and
    provides the `liquidsoap_version` attribute, updated every 10 seconds.
    It will also close PLAYOUT_CLIENT.
    """

    liquidsoap_version = "disconnected"

    def __init__(self, app: Quart | None = None):
        self.client = None
        self.app = None
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Quart) -> None:
        if "radiotomate_scheduler_watchdog" in app.extensions:
            raise RuntimeError(
                "radiotomate_scheduler_watchdog already registered on this app",
            )
        app.extensions["radiotomate_scheduler_watchdog"] = self
        self.app = app

        self.client = app.config["PLAYOUT_CLIENT"]
        app.before_serving(self.start_watchdog)
        app.after_serving(self.close_client)

    async def start_watchdog(self):
        self.app.add_background_task(self.watchdog)

    async def watchdog(self):
        """
        This is a watchdog that will send a GET request to the playout process'
        /live endpoint every 10 seconds, it will update liquidsoap_version and
        log accordingly.
        """
        previous_exception = None
        while True:
            try:
                r = await self.client.get("/version")
                if r.status_code == 200:
                    version = r.json()["version"]
                else:
                    version = "disconnected"
            except ShutdownError:
                break
            except Exception as e:
                if _log.isEnabledFor(logging.DEBUG):
                    if previous_exception and isinstance(e, previous_exception):
                        _log.debug("Caught again: %s", e)
                    else:
                        _log.exception(e)
                    previous_exception = e.__class__
                version = "disconnected"
            if Watchdog.liquidsoap_version != version:
                if version == "disconnected":
                    _log.warning("Could not connect to playout process")
                else:
                    _log.info(
                        "Connected to playout process - using liquidsoap %s",
                        version,
                    )
                Watchdog.liquidsoap_version = version
            try:
                await or_shutdown(asyncio.sleep(10))
            except (ShutdownError, asyncio.CancelledError):
                break

    async def close_client(self):
        await self.client.aclose()



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import asyncio
import logging

from quart import Blueprint, g, redirect, render_template, request, url_for
from quart_auth import current_user, login_user, logout_user
from werkzeug.exceptions import BadRequest, Unauthorized

from radiotomate.auth import RadiotomateAuth
from radiotomate.models import Session, User

_log = logging.getLogger(__name__)

blueprint = Blueprint("login", __name__, template_folder="templates")


@blueprint.get("/login")
async def index():
    return await render_template("login/index.jinja")


@blueprint.post("/login")
async def log_in():
    form = await request.form
    username = form.get("username")
    password = form.get("password")

    if not username:
        raise BadRequest("Please provide an username")
    if not password:
        raise BadRequest("Please provide a password")

    user = await User.check(g.dbsession, username, password)
    if not user:
        await asyncio.sleep(1)  # good luck, bruteforcers !
        raise Unauthorized("Incorrect username or password")
    session = Session(
        user_id=user.id,
        user_agent=request.headers.get("User-Agent"),
        latest_address=request.remote_addr,
    )
    g.dbsession.add(session)
    await g.dbsession.commit()

    login_user(RadiotomateAuth(session.id))
    return (
        "",
        302,
        {
            "HX-Redirect": url_for("home.index"),
        },
    )


@blueprint.route("/logout")
async def log_out():
    current_user.session.active = False
    await g.dbsession.commit()
    logout_user()
    return redirect(url_for("login.log_in"))



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Radiotomate DB entities, as SQLAlchemy mapped classes.

Those classes are used by the SQLAlchemy ORM to map the database tables to
Python objects. They can implement methods, but should not ``commit`` (at most,
``flush``).
"""

from radiotomate.models.autodj_slot import AutoDJSlot
from radiotomate.models.cart import Cart
from radiotomate.models.metadata_log import MetadataLog
from radiotomate.models.session import Session
from radiotomate.models.setting import Setting
from radiotomate.models.sound import Sound
from radiotomate.models.user import LoggedOutUser, User

__all__ = [
    "AutoDJSlot",
    "Cart",
    "LoggedOutUser",
    "MetadataLog",
    "Session",
    "Setting",
    "Sound",
    "User",
]



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import logging

import httpx
from quart import Blueprint, current_app, g, request

from radiotomate.auth import token_required
from radiotomate.models import MetadataLog

_log = logging.getLogger(__name__)

blueprint = Blueprint("metadata_log", __name__)


async def relay_metadata(raw_md):
    async with httpx.AsyncClient() as client:
        for target in current_app.config["RELAY_METADATA_TO"]:
            if not isinstance(target, dict):
                _log.warning("incorrect format for relay_to block: %s", str(target))
                continue
            url = target.get("url")
            data = target.get("add_field", {})
            data.update(raw_md)
            add_header = target.get("add_header")
            try:
                _ = await client.post(url, data=data, headers=add_header)
            except Exception as e:
                _log.error("Cannot relay metadata to %s : %s", url, e)


@blueprint.post("/metadata_log")
@token_required
async def post_metadata_log():
    raw_md = await request.get_json()

    if current_app.config["RELAY_METADATA_TO"]:
        # POST a copy of the dict because from_playout modifies it
        current_app.add_background_task(relay_metadata, dict(raw_md))

    md = await MetadataLog.from_playout(g.dbsession, raw_md)
    g.dbsession.add(md)
    await g.dbsession.commit()

    return "", 200



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
This module contains our extensions to the
[Quart-Auth](https://github.com/pgjones/quart-auth)
extension, that handles secure cookie authentication.
"""

from collections.abc import Awaitable, Callable
from datetime import datetime
from functools import wraps

from quart import current_app, g, redirect, request, url_for
from quart_auth import (
    AuthUser,
    Unauthorized,
    current_user,
    login_required,  # noqa: F401 so modules can import login_required from here
    logout_user,
)

from radiotomate.models import LoggedOutUser, Session, User


async def redirect_to_login(*_):
    """
    This is meant to be registered as an error handler of `Unauthorized` exceptions,
    and transforms them into a redirection to the login form.
    """
    return redirect(url_for("login.log_in"))


class RadiotomateAuth(AuthUser):
    """
    This preloads two attributes to Quart's `current_user` object:
    * `user`: the user object
    * `session`: the session object
    it also updates the session object, to track users' activity.

    It is invoked in a hook created by `radiotomate.interface_app.app_factory`.
    """

    def __init__(self, session_id):
        super().__init__(session_id)
        self.user: User = None
        self.session: Session = None

    async def preload_attributes(self, update_latest_action=True):
        if self.auth_id:
            self.session = await Session.from_id(g.dbsession, self.auth_id)
            if self.session:
                if not self.session.active:
                    logout_user()
                    raise Unauthorized("This session has been closed")
                self.user = self.session.user
                if self.session.latest_address != request.remote_addr:
                    self.session.latest_address = request.remote_addr
                if update_latest_action:
                    self.session.latest_action = datetime.now()
                await g.dbsession.commit()
            else:
                logout_user()
                raise Unauthorized("Invalid session")
        else:
            self.session = None
            self.user = LoggedOutUser()


def permission_required[**P](permission: str) -> Callable[P, Awaitable[object]]:
    """
    A decorator for Quart methods that checks current users' roles in Radiotomate.
    For example:

        @blueprint.delete("/carts/<int:cart_id>/sounds/<int:sound_id>")
        @permission_required("carts")
        async def delete_sound(cart_id: int, sound_id: int):
            ...

    This is a more specialized variant of QuartAuth's `@login_required`.
    """

    def decorator(
        func: Callable[P, Awaitable[object]],
    ) -> Callable[P, Awaitable[object]]:
        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs):
            if not await current_user.is_authenticated:
                raise Unauthorized()
            else:
                checker = getattr(current_user.user, "can_" + permission)
                if not checker():
                    raise Unauthorized()
                else:
                    return await current_app.ensure_async(func)(*args, **kwargs)

        return wrapper

    return decorator


def token_required[**P, T](
    func: Callable[P, Awaitable[T]],
) -> Callable[P, Awaitable[T]]:
    """A decorator that restricts route access between radiotomate's processes.

    This should be used to wrap a view function to ensure the request contains
    the correct token in the `X-Auth-Token` header.
    The value is configured in `playout_process_config.token` in the config file.

    Otherwise, raises `quart.exceptions.Unauthorized`.
    """

    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        if (
            not request.headers.get("X-Auth-Token")
            == current_app.config["PLAYOUT_TOKEN"]
        ):
            raise Unauthorized()
        else:
            return await current_app.ensure_async(func)(*args, **kwargs)

    return wrapper



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from __future__ import annotations

import logging
import secrets
from datetime import datetime

from sqlalchemy import Boolean, ForeignKey, Integer, String, select
from sqlalchemy.dialects.sqlite import DATETIME
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.orm import Session as ormSession

from radiotomate.db import Base

_log = logging.getLogger(__name__)


class Session(Base):
    """
    User session data - keeps track of who identified when, on which user-agent
    and from what IP address. Also notes when was the last action.

    Quart's session cookie only stores the ID of a line in this table, this
    ensure we can log out an user on the server side if their account is
    compromised.
    """

    __tablename__ = "sessions"
    id: Mapped[str] = mapped_column(
        String,
        primary_key=True,
        nullable=False,
        default=secrets.token_hex,
    )
    user_id: Mapped[int] = mapped_column(
        Integer,
        ForeignKey("users.id"),
        nullable=False,
        index=True,
    )
    active: Mapped[bool] = mapped_column(Boolean, nullable=False, default=True)
    user_agent: Mapped[str] = mapped_column(String)
    latest_address: Mapped[str] = mapped_column(String)
    created: Mapped[datetime] = mapped_column(
        DATETIME,
        default=datetime.now,
        nullable=False,
    )
    latest_action: Mapped[datetime] = mapped_column(
        DATETIME,
        default=datetime.now,
        onupdate=datetime.now,
        nullable=False,
    )

    user: Mapped[User] = relationship(lazy="joined", back_populates="sessions")  # noqa: F821

    @classmethod
    async def from_id(cls, db_session: ormSession, session_id: int) -> Session | None:
        return await db_session.scalar(select(Session).filter(Session.id == session_id))



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
The scheduler's ASGI app factory
"""

import logging
from asyncio import CancelledError
from collections.abc import Callable
from inspect import isawaitable
from pathlib import Path
from typing import Any

from apscheduler import AsyncScheduler, Job
from apscheduler.datastores.sqlalchemy import SQLAlchemyDataStore
from apscheduler.eventbrokers.local import LocalEventBroker
from apscheduler.executors.async_ import AsyncJobExecutor
from httpx import AsyncClient
from quart_auth import QuartAuth
from werkzeug.exceptions import HTTPException

import radiotomate.models
from radiotomate.auth import RadiotomateAuth
from radiotomate.beets import BeetsIntegration
from radiotomate.db import PathLike, QuartAlchemy
from radiotomate.quart import CustomQuart, ShutdownError
from radiotomate.scheduler import (
    analyzer,
    harbor,
    live,
    metadata_log,
    schedule,
    version,
)

_log = logging.getLogger(__name__)


class APSChedulerMiddleware:
    def __init__(self, asgi_app, scheduler):
        self.asgi_app = asgi_app
        self.scheduler = scheduler

    async def __call__(self, scope, receive, send):
        if scope["type"] == "lifespan":
            try:
                async with self.scheduler:
                    await self.scheduler.start_in_background()
                    return await self.asgi_app(scope, receive, send)
            except (ShutdownError, CancelledError):
                pass
        else:
            return await self.asgi_app(scope, receive, send)


class RadiotomateAPSChedulerDataStore(SQLAlchemyDataStore):
    def get_table_definitions(self):
        upstream_metadata = super().get_table_definitions()
        for table in upstream_metadata.tables.values():
            table.name = "apscheduler_" + table.name
        return upstream_metadata


class RadiotomateJobExecutor(AsyncJobExecutor):
    """
    like AsyncJobExecutor, but provides a DB session as the first argument and
    the current app as the second argument.
    """

    def __init__(self, app: CustomQuart, db: QuartAlchemy):
        self.app = app
        self.db = db

    async def run_job(self, func: Callable[..., Any], job: Job) -> Any:
        async with self.db.session() as session:
            retval = func(session, self.app, *job.args, **job.kwargs)
            if isawaitable(retval):
                retval = await retval
            return retval


def errors_as_text(exc: Exception) -> str:
    if not isinstance(exc, HTTPException):
        exc.code = 500
    return (str(exc), exc.code)


def app_factory(config: dict, beets: BeetsIntegration) -> CustomQuart:
    app = CustomQuart(__name__)
    app.config.from_mapping(
        {
            "ALCHEMY_ENGINE_CONFIG": config["db"],
            "DATA_ROOT": Path(config["data"]["root"]),
            "PLAYOUT_TOKEN": config["playout_process_config"]["token"],
            "PLAYOUT_CLIENT": AsyncClient(
                base_url="http://127.0.0.1:6833",
                headers={
                    "X-Auth-Token": config["playout_process_config"]["token"],
                },
            ),
            "RELAY_METADATA_TO": config.get("metadata_log", {}).get("relay_to", []),
        },
    )
    app.register_error_handler(Exception, errors_as_text)

    app.register_blueprint(analyzer.blueprint)
    app.register_blueprint(harbor.blueprint)
    app.register_blueprint(live.blueprint)
    app.register_blueprint(metadata_log.blueprint)
    app.register_blueprint(version.blueprint)
    app.register_blueprint(schedule.blueprint)

    PathLike.base_path = app.config["DATA_ROOT"]
    db = QuartAlchemy(app)

    beets.init_app(app)

    @app.before_serving
    async def load_settings():
        async with db.session() as session:
            dbsettings = await radiotomate.models.Setting.load_all(session)
            app.config.from_mapping(dbsettings)
        app.add_background_task(beets.background_analyzer, db)

    auth_manager = QuartAuth()
    auth_manager.user_class = RadiotomateAuth
    auth_manager.init_app(app)

    app.scheduler = AsyncScheduler(
        data_store=RadiotomateAPSChedulerDataStore(db.engine),
        event_broker=LocalEventBroker(),
        job_executors={"async": RadiotomateJobExecutor(app, db)},
        max_concurrent_jobs=1,
    )
    app.asgi_app = APSChedulerMiddleware(app.asgi_app, app.scheduler)

    return app



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
This module contains mostly HTML templates.

Additional templates filter functions are also defined in this file.
All must be reffered in `RADIOTOMATE_HELPERS` at the end of this file.
"""

from datetime import datetime, timedelta
from urllib.parse import urlparse

from markupsafe import Markup
from quart import request
from quart.templating import render_template_string
from quart_auth import current_user

from radiotomate.enums import CartMode, ScheduleMode


async def render_macro(path, macro, **kwargs) -> str:
    """
    Shortcut to render a macro. Parameters are
     - path to the file defining the macro
     - macro name
     - macro variables as keyword arguments

    As the path to the macro file is usually always the same in each module, we
    advise to create in each module a wrapper that hardcodes ``path``, so calls
    will be shorter.
    """
    return await render_template_string(
        f"""
        {{% from '{path}' import {macro} %}}
        {{{{ {macro}({", ".join(kwargs.keys())}) }}}}
        """,
        **kwargs,
    )


def in_blueprint(blueprint_name) -> bool:
    return request and request.blueprint and request.blueprint == blueprint_name


def date_span(dt: datetime):
    """
    Returns a formatted date in a span, and the complete datetime in a tooltip
    """
    return Markup(f"""<span title="{dt.isoformat()}">
            {dt.date().strftime("%d/%m/%Y")}
        </span>""")


def duration_span(seconds: int) -> str:
    """
    This generates a span that will show a formatted duration
    """
    delta = timedelta(seconds=seconds)
    return str(delta)


def icon(raw_unicode: str) -> str:
    """
    Wraps an unicode character in a clickable icon element hidden to screen readers.
    """
    return Markup(
        f"""<span class="icon is-clickable" aria-hidden="true">{raw_unicode}</span>"""
    )


def hostname_span(url: str) -> Markup:
    parsed = urlparse(url)
    if parsed:
        displayed = parsed.hostname
    else:
        displayed = url
    return Markup(f"""<span title="{Markup.escape(url)}">{displayed}</span>""")


RADIOTOMATE_HELPERS = {
    "in_blueprint": in_blueprint,
    "date_span": date_span,
    "hostname_span": hostname_span,
    "icon": icon,
    "transitions_delay_ms": 250,
    "current_user": current_user,
    "duration_span": duration_span,
    "CartMode": CartMode,
    "ScheduleMode": ScheduleMode,
}



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import logging

from quart import Blueprint, g, render_template

from radiotomate.auth import login_required
from radiotomate.models import MetadataLog

_log = logging.getLogger(__name__)

blueprint = Blueprint("history", __name__, template_folder="templates")


@blueprint.get("/history")
@login_required
async def index():
    history = await MetadataLog.get(g.dbsession, 200)
    return await render_template("history/index.jinja", history=history)



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

import click

from . import radiotomate_cli


@radiotomate_cli.group()
def users():
    """
    Users management commands
    """


@users.command()
def roles():
    """
    List role identifiers, that can be used with --role
    """
    from radiotomate.enums import Role

    for role in Role:
        click.echo(role)


ROLES_HELP = (
    "Set role. List available roles with the `roles` command. "
    "Repeat the option for each role you want to grant."
)


def _process_roles(user, role):
    from radiotomate.enums import Role

    roles = {}
    for r in role:
        if r not in Role:
            msg = (
                f"Role name '{r}' does not exist. "
                "Call `radiotomate user roles` to list possible values."
            )
            click.UsageError(msg)
        roles["can_" + r] = "true"
    user.update_roles(roles)


@users.command()
@click.argument("username")
@click.password_option()
@click.option("--admin", is_flag=True, help="Give admin rights to this account")
@click.option("--role", multiple=True, help=ROLES_HELP)
@click.pass_obj
def add(config, username, password, role, admin=False):
    """
    Create a radiotomate user account
    """
    import asyncio

    from radiotomate.db import QuartAlchemy
    from radiotomate.models import User

    db = QuartAlchemy()
    db.init_from_config(config["db"], {})
    user = User(username=username)
    user.update_password(password)

    if admin or "admin" in role:
        user.update_roles({"can_admin": "true"})
    else:
        _process_roles(user, role)

    async def do_add(session):
        async with session.begin() as dbsession:
            dbsession.add(user)

    asyncio.run(do_add(db.session))


@users.command()
@click.argument("username")
@click.password_option(
    prompt_required=False,
    help="If you prefer entering it interactively, "
    "write `--password` alone after the username.",
)
@click.option(
    "--role",
    multiple=True,
    help=ROLES_HELP + "Existing roles are untouched if the option is not used, "
    "but once one is provided other roles will be revoked.",
)
@click.pass_obj
def mod(config, username, password, role):
    """
    Change the password of a radiotomate user
    """
    import asyncio

    from radiotomate.db import QuartAlchemy
    from radiotomate.models import User

    db = QuartAlchemy()
    db.init_from_config(config["db"], {})

    async def do_mod(session):
        async with session.begin() as dbsession:
            user = await User.from_username(dbsession, username)
            if user:
                if password:
                    user.update_password(password)
                if role:
                    _process_roles(user, role)
            else:
                click.echo(f"User {username} not found")
                raise click.exceptions.Exit(1)

    asyncio.run(do_mod(db.session))


@users.command()
@click.argument("username")
@click.pass_obj
def rm(config, username):
    """
    Remove a radiotomate user account
    """
    import asyncio

    from radiotomate.db import QuartAlchemy
    from radiotomate.models import User

    db = QuartAlchemy()
    db.init_from_config(config["db"], {})

    async def do_mod(session):
        async with session.begin() as dbsession:
            user = await User.from_username(dbsession, username)
            if user:
                await dbsession.delete(user)
            else:
                click.echo(f"User {username} not found")
                raise click.exceptions.Exit(1)

    asyncio.run(do_mod(db.session))



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from sqlalchemy import text


async def upgrade(conn):
    #### Create the initial schema
    q = text("""CREATE TABLE IF NOT EXISTS settings (
                key TEXT PRIMARY KEY,
                value TEXT
                )
                """)
    await conn.execute(q)

    q = text("""CREATE TABLE users (
                id INTEGER PRIMARY KEY,
                username TEXT NOT NULL UNIQUE,
                password TEXT NOT NULL,
                notes TEXT NOT NULL DEFAULT '',
                created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                modified TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                roles JSON NOT NULL DEFAULT '{}'
                )
                """)
    await conn.execute(q)

    q = text("""CREATE TABLE sessions (
                id TEXT PRIMARY KEY,
                user_id INTEGER,
                user_agent TEXT,
                active BOOLEAN DEFAULT TRUE,
                latest_address TEXT,
                created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                latest_action TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                FOREIGN KEY(user_id) REFERENCES users(id) ON DELETE CASCADE
                )
                """)
    await conn.execute(q)
    q = text("""CREATE INDEX sessions_user_id_idx ON sessions(user_id)""")
    await conn.execute(q)

    q = text("""CREATE TABLE carts (
                id INTEGER PRIMARY KEY,
                title TEXT NOT NULL UNIQUE,
                path TEXT,
                url TEXT,
                average_duration INTEGER NOT NULL DEFAULT 0,
                max_duration INTEGER,
                notes TEXT NOT NULL DEFAULT '',
                mode TEXT NOT NULL DEFAULT 'playlist',
                schedule_mode TEXT NOT NULL DEFAULT 'timed',
                schedule_correct BOOLEAN NOT NULL DEFAULT FALSE,
                schedule_year TEXT NOT NULL DEFAULT 'timed',
                schedule_month TEXT NOT NULL DEFAULT 'timed',
                schedule_day TEXT NOT NULL DEFAULT 'timed',
                schedule_week TEXT NOT NULL DEFAULT 'timed',
                schedule_day_of_week TEXT NOT NULL DEFAULT 'timed',
                schedule_hour TEXT NOT NULL DEFAULT 'timed',
                schedule_minute TEXT NOT NULL DEFAULT 'timed',
                schedule_second TEXT NOT NULL DEFAULT 'timed',
                created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                modified TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP
                )
                """)
    await conn.execute(q)

    q = text("""CREATE TABLE sounds (
                id INTEGER PRIMARY KEY,
                cart_id INTEGER NOT NULL,
                rank INTEGER NOT NULL DEFAULT 1,
                path TEXT NOT NULL,
                duration INTEGER NOT NULL DEFAULT 0,
                gain FLOAT,
                peak FLOAT,
                title TEXT NOT NULL,
                active BOOLEAN DEFAULT TRUE,
                uploader_id INTEGER,
                created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                last_played TIMESTAMP,
                FOREIGN KEY(cart_id) REFERENCES carts(id) ON DELETE CASCADE
                )
                """)
    await conn.execute(q)
    q = text("""CREATE INDEX sounds_cart_id_idx ON sounds(cart_id)""")
    await conn.execute(q)

    q = text("""CREATE TABLE metadata_log (
                id INTEGER PRIMARY KEY,
                on_air TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                source TEXT,
                source_url TEXT,
                cart_id INTEGER,
                artist TEXT,
                title TEXT,
                album TEXT,
                extra JSON NOT NULL DEFAULT '{}',
                FOREIGN KEY(cart_id) REFERENCES carts(id) ON DELETE SET NULL
                )
                """)
    await conn.execute(q)
    q = text("""CREATE INDEX metadata_log_on_air_idx ON metadata_log(on_air)""")
    await conn.execute(q)

    q = text("""CREATE TABLE autodj_slots (
                id INTEGER PRIMARY KEY,
                day_of_week INTEGER NOT NULL 
                    CHECK (0 <= day_of_week AND day_of_week <= 6),
                minute INTEGER NOT NULL 
                    CHECK (0 <= day_of_week AND day_of_week < 1440),
                color TEXT NOT NULL DEFAULT '',
                title TEXT NOT NULL DEFAULT '',
                constraints JSON NOT NULL DEFAULT '{}',
                created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                modified TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP
                )""")
    await conn.execute(q)

    q = text("""CREATE UNIQUE INDEX ix_autodj_slots_uq_day_minute
             ON autodj_slots (day_of_week, minute)""")
    await conn.execute(q)

    q = text("""INSERT INTO autodj_slots (day_of_week, minute, color)
             VALUES (:d, 0, '#efefef')""")
    for d in range(7):
        await conn.execute(q, {"d": d})

    #### Create basic settings
    q = text("select value from settings where key=:key")
    result = await conn.execute(q, {"key": "SCHEMA_VERSION"})
    if (
        result.scalars().first() is None
    ):  # the opposite may happen if the install was downgraded to v0...
        # which is weird, but happens when developing
        q = text("""INSERT INTO settings (key, value) VALUES (:key, :value)""")
        await conn.execute(q, {"key": "SCHEMA_VERSION", "value": "1"})

    default_settings = {
        "INTERFACE_NAME": "Radiotomate",
    }
    q = text("""INSERT INTO settings (key, value) VALUES (:key, :value)""")
    for key, value in default_settings.items():
        await conn.execute(q, {"value": value, "key": key})


async def downgrade(conn):
    q = text("DROP INDEX ix_autodj_slots_uq_day_minute")
    await conn.execute(q)

    q = text("DROP TABLE autodj_slots")
    await conn.execute(q)

    q = text("DROP TABLE sessions")
    await conn.execute(q)

    q = text("DROP TABLE users")
    await conn.execute(q)

    q = text("DROP TABLE sounds")
    await conn.execute(q)

    q = text("DROP TABLE carts")
    await conn.execute(q)

    # don't drop settings because the update system will attempt to update it
    # we can truncate it, instead
    q = text("""DELETE FROM settings""")
    await conn.execute(q)
    q = text("""INSERT INTO settings (key, value) VALUES (:key, :value)""")
    await conn.execute(q, {"value": "0", "key": "SCHEMA_VERSION"})



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

from __future__ import annotations

import contextlib
from pathlib import Path
from typing import TYPE_CHECKING

from quart import Quart, current_app, g, request, request_tearing_down
from sqlalchemy import Dialect, String, event
from sqlalchemy.ext.asyncio import (
    AsyncAttrs,
    async_engine_from_config,
    async_sessionmaker,
)
from sqlalchemy.orm import DeclarativeBase, registry
from sqlalchemy.pool import AsyncAdaptedQueuePool
from sqlalchemy.types import TypeDecorator

if TYPE_CHECKING:
    from sqlalchemy.engine.interfaces import DBAPICursor


class PathLike(TypeDecorator):
    """
    Allows to map columns as `Path` objects containing absolute paths, while
    storing strings relative to `base_path`.

    This reduces the DB size and let the user move the `DATA_ROOT` folder, they
    should only update the configuration. However applications must copy
    `base_path` from their configuration when starting.
    """

    impl = String
    base_path: Path = None

    def process_bind_param(self, value: Path | None, dialect: Dialect) -> str:
        """Convert an `Path` value to a string for the database."""
        if value:
            if not self.base_path:
                raise RuntimeError("Application factory must set PathLike.base_path")
            return str(value.relative_to(self.base_path))
        return value

    def process_result_value(
        self,
        value: str | None,
        dialect: Dialect,
    ) -> Path | None:
        """Restore a string from the database to a `Path`."""
        if value is not None:
            if not self.base_path:
                raise RuntimeError("Application factory must set PathLike.base_path")
            return self.base_path / value
        return value


class Base(AsyncAttrs, DeclarativeBase):
    registry = registry(
        type_annotation_map={
            Path: PathLike,
        },
    )


class QuartAlchemy:
    """
    This expect an app whose config contain a `ALCHEMY_ENGINE_CONFIG`
    key with a SQLAlchemy engine configuration: a dict that should
    contain at least a `url` key. For other options see SQLAlchemy's
    `async_engine_from_config`.
    The config may also contain a `ALCHEMY_SESSION_CONFIG`, associated
    to a dict matching `async_sessionmaker`'s parameters (we'll set
    `expire_on_commit=False` by default).

    Activate the extension with
    ```
    db = QuartAlchemy(app)
    ```

    Then, you'll have a per-request session object in the `g` object.
    Don't forget to `commit()`:
    ```
    g.dbsession.add(user)
    await g.dbsession.commit()
    ```

    This session is not open when serving from the `static` route.

    Outside of requests contexts, use its `session` attribute to get an
    `AsyncSession` object. For example:

    ```
    async with db.session() as session:
        session.add(SomeObject(data="object"))
        session.add(SomeOtherObject(name="other object"))
        await session.commit()
    ```

    The context manager can also handle the transaction with:

    ```
    async with db.session.begin() as session:
        session.add(SomeObject(data="object"))
    ```

    You may also access the engine directly, typically to create tables:

    ```
    async with db.engine.begin() as conn:
        await conn.run_sync(Base.metadata.drop_all)
        await conn.run_sync(Base.metadata.create_all)
    ```

    As we're using async drivers, even when running in a synchronous context
    you'll need to wrap such code in a an asynchronous function:

    ```
    async def create_all(engine):
        async with engine.connect() as conn:
            conn.run_sync(Base.metadata.create_all)
    asyncio.run(create_all(db.engine))
    ```

    """

    EXTENSION_NAME = "sqlalchemy"

    def __init__(self, app: Quart | None = None):
        self.engine = None
        self.session = None
        self.engine_config = None
        self.session_config = None
        self.in_memory = False
        if app is not None:
            self.init_app(app)

    @classmethod
    def get(cls) -> QuartAlchemy:
        """
        Find the instance wrapped in ``current_app``
        """
        return current_app.extensions[cls.EXTENSION_NAME]

    def init_app(self, app: Quart) -> None:
        if self.EXTENSION_NAME in app.extensions:
            raise RuntimeError(
                "An sqlalchemy extension has already been registered on this app",
            )
        app.extensions[self.EXTENSION_NAME] = self

        self.init_from_config(
            app.config.get("ALCHEMY_ENGINE_CONFIG"),
            app.config.get("ALCHEMY_SESSION_CONFIG", {}),
        )
        app.after_serving(self.after_serving)
        app.before_request(self.before_request)
        request_tearing_down.connect(self.teardown_request, app)

    @staticmethod
    def sqlite_pragmas(cursor: DBAPICursor) -> None:
        """
        This ensures we use the same PRAGMA statement in all contexts:
         * when initilizing SQLAlchemy
         * when testing

        This list is inspired from https://gcollazo.com/optimal-sqlite-settings-for-django/
        and linked articles
        """
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.execute("PRAGMA journal_mode=WAL")
        cursor.execute("PRAGMA synchronous=NORMAL")
        cursor.execute("PRAGMA busy_timeout=5000")  # 5s
        cursor.execute("PRAGMA temp_store=MEMORY")
        cursor.execute("PRAGMA mmap_size=134217728")  # 128MB
        cursor.execute("PRAGMA cache_size=10000")  # 10k pages

    def init_from_config(self, engine_config: dict, session_config: dict):
        """
        This really starts the SQLAlchemy engine and ensures that ``self.session()``
        will return something useful later on.
        When not binding to a Quart app, calling this method is enough.
        """
        self.engine_config = {
            "pool_size": 10,
            "max_overflow": 10,
            "pool_recycle": 3600,
            "poolclass": AsyncAdaptedQueuePool,
        }
        self.engine_config.update(engine_config)
        self.engine = async_engine_from_config(self.engine_config, prefix="")

        @event.listens_for(self.engine.sync_engine, "connect")
        def do_connect(dbapi_connection, connection_record):
            # cf. https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#enabling-non-legacy-sqlite-transactional-modes-with-the-sqlite3-or-aiosqlite-driver
            # we should put "connect_args": {"autocommit": False}, in engine_config,
            # but that would trigger errors in sqlite_pragmas like
            # "cannot change into wal mode from within a transaction"
            # so we're left with the legacy mode:
            dbapi_connection.isolation_level = None
            # this can't be changed until AsyncAdapt_aiosqlite_connection has no
            # `autocommit` attribute - if that ever happens, we could apply the
            # documented trick that temporarly sets autocommit while calling
            # sqlite_pragmas

            cursor = dbapi_connection.cursor()
            self.sqlite_pragmas(cursor)
            cursor.close()

        self.session_config = {
            "expire_on_commit": False,
            "bind": self.engine,
        }
        self.session_config.update(session_config)
        self.session = async_sessionmaker(**self.session_config)

    async def before_request(self) -> None:
        if request.endpoint != "static":
            g.dbsession = self.session()

    async def teardown_request(self, sender, **extra) -> None:
        # extra may include "exc" if an exception occurred

        # AttributeError happens when serving the static endpoint
        with contextlib.suppress(AttributeError):
            await g.dbsession.close()

    async def after_serving(self) -> None:
        if self.engine:
            await self.engine.dispose()



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
DB Schema upgrade system.

Each version of the schema is in a sub-module, named `vX.py`, where `X` is the version
number. Each sub-module must define functions `upgrade(conn)` and `downgrade(conn)`,
that will be called with a SQLAlchemy connection object.

"""

import click

from .. import radiotomate_cli


async def do_update(engine, target_version=None):
    import sys
    from importlib import import_module
    from pkgutil import iter_modules

    from sqlalchemy import text
    from sqlalchemy.exc import OperationalError

    async with engine.connect() as conn:
        try:
            q = text("select value from settings where key=:key")
            result = await conn.execute(q, {"key": "SCHEMA_VERSION"})
            current_version = int(result.scalars().first())
        except OperationalError:  # when table does not exist
            current_version = 0

        latest_version = 0
        updates_module = sys.modules[__name__]
        for modinfo in iter_modules(updates_module.__path__):
            name = modinfo.name
            if name.startswith("v"):
                version = int(name[1:])
                latest_version = max(version, latest_version)

        if target_version is None:
            target_version = latest_version

        if target_version > latest_version:
            click.secho(
                f"Version {target_version} does not exists",
                fg="red",
                bold=True,
            )
            raise click.exceptions.Exit(1)

        if current_version < target_version:
            for i in range(current_version + 1, target_version + 1):
                click.echo(f"Applying schema change #{i}... ", nl=False)
                module = import_module(f"radiotomate.commands.update.v{i}")
                await module.upgrade(conn)
                q = text("update settings set value=:value where key=:key")
                await conn.execute(q, {"value": str(i), "key": "SCHEMA_VERSION"})
                await conn.commit()
                click.secho("OK", fg="green", bold=True)
        elif current_version > target_version:
            for i in range(current_version, target_version, -1):
                click.echo(f"Reverting schema change # {i - 1}... ", nl=False)
                module = import_module(f"radiotomate.commands.update.v{i}")
                await module.downgrade(conn)
                q = text("update settings set value=:value where key=:key")
                await conn.execute(q, {"value": str(i - 1), "key": "SCHEMA_VERSION"})
                await conn.commit()
                click.secho("OK", fg="green", bold=True)
        else:
            click.secho("DB is up-to-date", fg="green", bold=True)


@radiotomate_cli.command()
@click.option(
    "--target-version",
    help="Target version (may be used to downgrade)",
    type=int,
)
@click.pass_obj
def update(config, target_version=None):
    """
    Update a Radiotomate installation
    """
    import asyncio

    # do *not* import radiotomate.models from here, because their tables may not exist
    # yet. though, we use db.init_from_config() in order to have the same PRAGMAs
    # and settings as the app
    from radiotomate.db import QuartAlchemy

    db = QuartAlchemy()
    db.init_from_config(config["db"], {})
    asyncio.run(do_update(db.engine, target_version))
    asyncio.run(db.after_serving())



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Radiotomate instance creation commands
======================================

Those commands can create a configuration file, a database, and also install
system services.
"""

from pathlib import Path

import click

from .. import load_config, radiotomate_cli
from ..update import update


@radiotomate_cli.command()
@click.option(
    "--no-revert-on-failure",
    is_flag=True,
    help="If you want to develop/debug this installer",
)
@click.option(
    "-d",
    "--data-root",
    help="Path to radiotomate's data folder",
    type=click.Path(
        allow_dash=False,
        exists=False,
        resolve_path=True,
        path_type=Path,
    ),
)
@click.pass_context
def install(ctx, no_revert_on_failure, data_root: click.Path | None = None):
    """
    Set-up a new Radiotomate installation
    """
    from .installer import Installer

    installer = None
    try:
        click.secho("\nWelcome to the Radiotomate installer!", fg="green", bold=True)
        click.echo(
            "\nHit Ctrl+C to abort. Press Enter "
            "to use the default value (shown between square brackets).\n"
            "",
        )

        if not data_root:
            default = str(Path("./radio_data").absolute())
            data_root = Path(
                click.prompt(
                    "In which folder should Radiotomate store all its data?",
                    default=str(default),
                )
            )

        installer = Installer(data_root.absolute())
        if installer.instance_exists():
            click.secho(
                "Note: it seems an instance already exists in this folder, "
                "we'll only try to update it",
                bold=True,
                fg="red",
            )
        else:
            installer.install()

        ctx.params["config_path"] = installer.path_yaml
        ctx.obj = load_config(installer.path_yaml)
        ctx.invoke(update)

    except Exception:
        if installer and not no_revert_on_failure:
            installer.revert()
        raise

    installer.recap()


@radiotomate_cli.command()
@click.option(
    "--quiet",
    is_flag=True,
    help="do not print the recap when done - useful for testing",
)
@click.pass_context
def develop(ctx, quiet):
    """
    Set-up a new Radiotomate installation for development or testing (create only the
    database and the configuration file)
    """
    from .installer import Installer

    target_path = Path("./radio_data").absolute()
    installer = Installer(target_path)
    if installer.instance_exists():
        raise click.UsageError(
            f"It seems a development install already exists in {target_path}"
        )
    installer.install(dev=True)

    ctx.params["config_path"] = installer.path_yaml
    ctx.obj = load_config(installer.path_yaml, verbose=True)
    ctx.invoke(update)

    if not quiet:
        click.secho("\nRadiotomate is installed for development", fg="green", bold=True)
        click.echo("We advise you create an admin user using:")
        click.echo(
            f"    radiotomate -c {installer.path_yaml} useradd --admin [USERNAME]"
        )
        click.echo(
            "Then, you can call ./dev.sh to start the system: log in, create a "
            "jingles cart (put at least one sound), and add some music."
        )



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
This module defines endpoints related to the harbor "stream" input.
"""

import logging

from quart import Blueprint, g, request
from werkzeug.exceptions import BadRequest, Unauthorized

from radiotomate.auth import token_required
from radiotomate.models import User

_log = logging.getLogger(__name__)

blueprint = Blueprint("harbor", __name__)


@blueprint.post("/can_stream")
@token_required
async def can_stream():
    posted = await request.get_json()
    username = posted.get("user")
    password = posted.get("password")

    if not username:
        raise BadRequest("Please provide an username")
    if not password:
        raise BadRequest("Please provide a password")

    user = await User.check(g.dbsession, username, password)
    if not user:
        raise Unauthorized("Incorrect username or password")
    if not user.can_stream():
        raise Unauthorized("User is not authorized to stream")

    return ""



# SPDX-FileCopyrightText: 2026 Martin Kirchgessner
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Scheduler API
=============

Orders sent to the scheduler app are wrapped in an object that can be mocked
for tests or the demo mode, where the scheduler process does not exist.
"""

from __future__ import annotations

import json
import logging
import random
from asyncio import sleep
from datetime import datetime
from typing import TYPE_CHECKING

import httpx

from radiotomate.db import QuartAlchemy
from radiotomate.models import Sound
from radiotomate.quart import ShutdownError, or_shutdown

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

    from radiotomate.quart import CustomQuart

_log = logging.getLogger(__name__)


class Scheduler:
    """
    This is a singleton: use `Scheduler.get()`.
    It should be instantiated with `Scheduler.init(config, app)` when starting
    the application.
    """

    _instance = None

    client: httpx.AsyncClient = None

    @classmethod
    def get(cls) -> Scheduler:
        if cls._instance is None:
            raise RuntimeError("Scheduler is not initialized")
        return cls._instance

    @classmethod
    def init(cls, config, app: CustomQuart, demo=False) -> Scheduler:
        if cls._instance is not None:
            raise RuntimeError("init has already been called")
        if demo:
            implementation = SchedulerDemo
        else:
            implementation = cls
        cls._instance = implementation(config, app)
        app.after_serving(cls.after_serving)
        return cls._instance

    @classmethod
    async def after_serving(cls) -> None:
        if cls._instance:
            await cls._instance.client.aclose()

    def __init__(self, config, app: CustomQuart):
        self.client = httpx.AsyncClient(
            base_url="http://127.0.0.1:6822",
            headers={
                "X-Auth-Token": config["playout_process_config"]["token"],
            },
        )

    async def skip(self):
        await self.client.delete("/live")

    async def live(self) -> AsyncGenerator[dict]:
        sleeptime = 10
        while True:
            try:
                if sleeptime:
                    await or_shutdown(sleep(sleeptime))
                    sleeptime = 0
                async with self.client.stream("GET", "/live") as response:
                    async for line in response.aiter_bytes():
                        if response.status_code == 200 and line.startswith(
                            b"event: message\ndata: ",
                        ):
                            yield json.loads(line[21:])
            except ShutdownError:
                break
            except httpx.ReadTimeout:
                # avoid the verbose _log.exception when playout process is not started
                _log.debug(
                    "No live info received from scheduler, is the player started ?",
                )
                sleeptime = 10
            except httpx.ConnectError:
                # avoids the verbose _log.exception when working only on the interface
                _log.debug(
                    "Cannot connect to %s - retrying in 10 seconds",
                    self.client.base_url,
                )
                sleeptime = 10
            except Exception as e:
                if _log.isEnabledFor(logging.DEBUG):
                    _log.exception(e)
                sleeptime = 3

    async def update_schedule(self, cart_id: int):
        _log.debug("Updating schedule for cart %d", cart_id)
        result = await self.client.put(f"/schedule/{cart_id}")
        if result.status_code != 200:
            _log.error(
                "Error while updating schedule: %s %s",
                result.status_code,
                result.text,
            )

    async def queue_analysis(self, sound_ids: list[int]):
        _log.debug("Queueing sounds for analysis: %r", sound_ids)
        result = await self.client.post(
            "/analyzer",
            json={
                "sound_ids": sound_ids,
            },
        )
        if result.status_code != 200:
            _log.error(
                "Queueing sounds (ids: %r) for analysis: %s %s",
                sound_ids,
                result.status_code,
                result.text,
            )


WORDS = list(
    {
        # From Tetsuo Kogawa' radioart manifesto https://radioart.jp
        w
        for w in """
    What is radioart? Who is radioart? Popular meaning of “radio” has been a
    receiving tool of radio signals There is and can be radioart using such a
    tool More positively radioart would be involved in wireless transmission
    However such a transmission remains in the function of broadcasting
    Radio station broadcasts Broadcast means 'cast broadly'
    Sometimes broadcasting is done not so broadly It is called "narrowcasting"
    But it still does cast Broadcasting has been seeking for more and more
    broader range of transmission toward nation-wide worldwide and space-wide
    (satellite) broadcasting""".split()  # noqa: SIM905
        if len(w) >= 3
    },
)


def artistic_generator():
    return " ".join([WORDS[random.randint(0, len(WORDS) - 1)] for i in range(2)])


class SchedulerDemo(Scheduler):
    """
    Demo/tests mode: in that case there is no scheduler process, so we override
    everything.

    In that case the `client` instance actualy points to the interface app,
    so this can emulate playing metadata.
    """

    TRACK_LENGTH = 180

    def __init__(self, config, app: CustomQuart):
        self.client = httpx.AsyncClient(
            base_url="http://127.0.0.1:6811",
            headers={
                "X-Auth-Token": config["playout_process_config"]["token"],
            },
        )
        self._started_at = datetime.now().replace(microsecond=0)
        self._on_air = datetime.now().replace(microsecond=0)
        self._i = 0
        self._metadata = self.generate_metadata()
        self.app = app

    async def live(self) -> AsyncGenerator[dict]:
        while True:
            elapsed = (datetime.now() - self._on_air).total_seconds()
            remaining = self.TRACK_LENGTH - elapsed
            if remaining < 0.0:
                await self.skip()
                remaining = 0.0
            md = self._metadata
            md["uptime"] = self.uptime()
            md["time"] = datetime.now().replace(microsecond=0).isoformat()
            md["remaining"] = str(remaining)
            md["elapsed"] = str(elapsed)
            yield md
            try:
                await or_shutdown(sleep(1))
            except ShutdownError:
                break

    async def skip(self):
        self._on_air = datetime.now().replace(microsecond=0)
        self._metadata = self.generate_metadata()

    def uptime(self) -> str:
        uptime = datetime.now().replace(microsecond=0) - self._started_at
        return str(uptime)

    def generate_metadata(self) -> dict:
        """
        generates something different each call
        """
        sources = ["test", "live", "unkown_sound_source1231"]
        albums = ["Radiotomate Rocks", "Better hygiene with Liquidsoap"]

        self._i += 1
        title = artistic_generator()
        artist = artistic_generator()
        album = albums[self._i % len(albums)]
        filename = (
            f"/home/radio/Music/{album}/"
            + artist.replace(" ", "_")
            + "-"
            + title.replace(" ", "_")
            + ".flac"
        )
        return {
            "artist": artist,
            "title": title,
            "source": sources[self._i % len(sources)],
            "status": "simulating",
            "initial_uri": filename,
            "album": album,
            "editor": "Pytest",
            "year": self._on_air.year - (self._i % 10),
            "tracknumber": str(self._i % 8),
            "uptime": self.uptime(),
            "on_air": self._on_air.isoformat(),
        }

    async def update_schedule(self, cart_id: int):
        _log.debug("Fake-update of schedule for cart %d", cart_id)

    async def queue_analysis(self, sound_ids: list[int]):
        _log.debug("Queueing sounds for analysis: %r", sound_ids)
        self.app.add_background_task(self._fake_analysis, sound_ids)

    async def _fake_analysis(self, sound_ids):
        await sleep(0.1)
        db = QuartAlchemy.get()
        async with db.session() as session:
            for sound_id in sound_ids:
                sound = await Sound.from_id(session, sound_id)
                if sound:
                    sound.gain = -1
                    sound.peak = 0.5
            await session.commit()
